import datetime
import itertools
import math
from typing import Dict, List, Tuple, Union

import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from webviz_config import WebvizConfigTheme

from ...._types import ChartType
from ...._utils import EnsembleWellAnalysisData


class WellOverviewFigure:
    def __init__(
        self,
        ensembles: List[str],
        data_models: Dict[str, EnsembleWellAnalysisData],
        sumvec: str,
        prod_from_date: Union[datetime.datetime, None],
        prod_until_date: Union[datetime.datetime, None],
        charttype: ChartType,
        wells: List[str],
        theme: WebvizConfigTheme,
    ) -> None:
        # pylint: disable=too-many-arguments
        self._ensembles = ensembles
        self._data_models = data_models
        self._sumvec = sumvec
        self._prod_from_date = prod_from_date
        self._prod_until_date = prod_until_date
        self._charttype = charttype
        self._wells = wells
        self._colors = theme.plotly_theme["layout"]["colorway"]
        self._rows, self._cols = self.get_subplot_dim()
        spec_type = "scatter" if self._charttype == ChartType.AREA else self._charttype
        subplot_titles = None if self._charttype == ChartType.BAR else self._ensembles

        self._figure = make_subplots(
            rows=self._rows,
            cols=self._cols,
            specs=[
                [{"type": spec_type} for _ in range(self._cols)]
                for _ in range(self._rows)
            ],
            subplot_titles=subplot_titles,
        )

        self._add_traces()

    @property
    def figure(self) -> go.Figure:
        return self._figure

    def get_subplot_dim(self) -> Tuple[int, int]:
        """Returns the subplot dimensions of the currently selected chart type.
        Pie charts have two columns, while the other has one. Area charts have one
        row per ensemble.
        """
        number_of_ens = len(self._ensembles)
        if self._charttype == ChartType.BAR:
            return 2, 1
        if self._charttype == ChartType.PIE:
            return max(math.ceil(number_of_ens / 2), 2), 2
        if self._charttype == ChartType.AREA:
            return number_of_ens, 1
        raise ValueError(f"Chart type: {self._charttype.value} not implemented")

    def _get_ensemble_charttype_data(self, ensemble: str) -> pd.DataFrame:
        """Returns a dataframe with summary data on the form needed for the
        different chart types.
        """
        if self._charttype in [ChartType.BAR, ChartType.PIE]:
            df = self._data_models[ensemble].get_dataframe_melted(
                well_sumvec=self._sumvec,
                prod_from_date=self._prod_from_date,
                prod_until_date=self._prod_until_date,
            )
            df = df[df["WELL"].isin(self._wells)]
            df_mean = df.groupby("WELL").mean().reset_index()
            return df_mean[df_mean[self._sumvec] > 0]

        # else chart type == area
        df = self._data_models[ensemble].get_summary_data(
            well_sumvec=self._sumvec,
            prod_from_date=self._prod_from_date,
            prod_until_date=self._prod_until_date,
        )
        return df.groupby("DATE").mean().reset_index()

    def _add_traces(self) -> None:
        """Add all traces for the currently selected chart type."""
        wells_in_legend = []

        for i, ensemble in enumerate(self._ensembles):
            df = self._get_ensemble_charttype_data(ensemble)

            if self._charttype == ChartType.PIE:
                self._figure.add_trace(
                    go.Pie(
                        values=df[self._sumvec],
                        labels=df["WELL"],
                        marker_colors=self._colors,
                        textposition="inside",
                        texttemplate="%{label}",
                    ),
                    row=i // 2 + 1,
                    col=i % 2 + 1,
                )

            elif self._charttype == ChartType.BAR:
                trace = {
                    "x": df["WELL"],
                    "y": df[self._sumvec],
                    "orientation": "v",
                    "type": "bar",
                    "name": ensemble,
                    "marker": {"color": self._colors[i]},
                    "text": df[self._sumvec],
                    "textposition": "none",
                    "texttemplate": "%{text:.2s}",
                }

                self._figure.add_trace(
                    trace,
                    row=1,
                    col=1,
                )
            elif self._charttype == ChartType.AREA:
                color_iterator = itertools.cycle(self._colors)

                for well in self._data_models[ensemble].wells:
                    if well in self._wells:
                        showlegend = False
                        if well not in wells_in_legend:
                            showlegend = True
                            wells_in_legend.append(well)

                        self._figure.add_trace(
                            go.Scatter(
                                x=df["DATE"],
                                y=df[f"{self._sumvec}:{well}"],
                                hoverinfo="text+x+y",
                                hoveron="fills",
                                mode="lines",
                                stackgroup="one",
                                name=well,
                                line=dict(width=0.1, color=next(color_iterator)),
                                legendgroup="Wells",
                                showlegend=showlegend,
                            ),
                            row=i + 1,
                            col=1,
                        )


def format_well_overview_figure(
    figure: go.Figure,
    charttype: ChartType,
    settings: List[str],
    sumvec: str,
    prod_from_date: Union[str, None],
    prod_until_date: Union[str, None],
) -> go.Figure:
    """This function formate the well overview figure. The reason for keeping this
    function outside the figure class is that we can update the figure formatting
    without generating a new WellOverviewFigure object and reloading the data. It can
    be applied directly to the current state of the figure dict if only formatting
    settings are changed. See in the well_overview_callbacks how it is used.
    """

    if charttype == ChartType.PIE:
        figure.update_traces(
            texttemplate=(
                "%{label}<br>%{value:.2s}"
                if "show_prod_text" in settings
                else "%{label}"
            )
        )

    elif charttype == ChartType.BAR:
        figure.update_layout(
            barmode=("overlay" if "overlay_bars" in settings else "group")
        )
        figure.update_traces(
            textposition=("auto" if "show_prod_text" in settings else "none")
        )

    # These are valid for all chart types
    figure.update_layout(
        template=("plotly_white" if "white_background" in settings else "plotly")
    )

    # Make title
    phase = {"WOPT": "Oil", "WGPT": "Gas", "WWPT": "Water"}[sumvec]
    title = f"Cumulative Well {phase} Production (Sm3)"
    if prod_from_date is not None:
        title += f" from {prod_from_date}"
    if prod_until_date is not None:
        title += f" until {prod_until_date}"

    figure.update(
        layout_title_text=title,
        layout_showlegend=("legend" in settings),
    )
    return figure



# pylint: disable=no-name-in-module
from webviz_config.plugins import HistoryMatch


def test_history_match(dash_duo, app, testdata_folder, shared_settings) -> None:

    plugin = HistoryMatch(
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        observation_file=testdata_folder
        / "01_drogon_ahm"
        / "share"
        / "observations"
        / "tables"
        / "ert_observations.yml",
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



# pylint: disable=no-name-in-module
from webviz_config.plugins import ParameterCorrelation
from webviz_config.testing import WebvizComposite


def test_parameter_correlation(_webviz_duo: WebvizComposite, shared_settings) -> None:

    parameter_correlation = ParameterCorrelation(
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
    )

    _webviz_duo.start_server(parameter_correlation)

    _webviz_duo.toggle_webviz_settings_drawer()
    _webviz_duo.toggle_webviz_settings_group(
        parameter_correlation.view("paracorr").settings_group_unique_id("settings")
    )
    # Using str literals directly, not IDs from the plugin as intended because
    # the run test did not accept the imports

    my_component_id = _webviz_duo.view_settings_group_unique_component_id(
        "paracorr", "settings", "shared-ensemble"
    )
    _webviz_duo.wait_for_contains_text(my_component_id, "iter-0")
    assert _webviz_duo.get_logs() == []



from dataclasses import asdict
from typing import Dict, List, Optional, Tuple

import jwt
import numpy as np
from dash import Input, Output, State, callback, html, no_update
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewABC

from webviz_subsurface._providers.ensemble_grid_provider import (
    CellFilter,
    EnsembleGridProvider,
    GridVizService,
    PropertySpec,
)

from ..._layout_elements import ElementIds
from ..._types import PROPERTYTYPE
from .settings import ColorScale, DataSettings, GridFilter
from .view_elements._vtk_view_3d_element import VTKView3D


class View3D(ViewABC):
    class Ids(StrEnum):
        VTKVIEW = "vtkview"
        DATASELECTORS = "data-selectors"
        GRIDFILTER = "grid-filter"
        COLORSCALE = "color-scale"

    def __init__(
        self,
        grid_provider: EnsembleGridProvider,
        grid_viz_service: GridVizService,
        initial_grid_filter: Dict[str, int],
    ) -> None:
        super().__init__("Grid View")
        self.grid_provider = grid_provider
        self.grid_viz_service = grid_viz_service
        self.vtk_view_3d = VTKView3D()
        self.add_view_element(self.vtk_view_3d, View3D.Ids.VTKVIEW)
        self.add_settings_group(
            DataSettings(grid_provider=grid_provider), View3D.Ids.DATASELECTORS
        )
        self.add_settings_group(
            GridFilter(
                grid_provider=grid_provider, initial_grid_filter=initial_grid_filter
            ),
            View3D.Ids.GRIDFILTER,
        )
        self.add_settings_group(ColorScale(), View3D.Ids.COLORSCALE)

    def _view_id(self, component_id: str) -> str:
        return (
            self.view_element(View3D.Ids.VTKVIEW)
            .component_unique_id(component_id)
            .to_string()
        )

    def _data_settings_id(self, component_id: str) -> str:
        return (
            self.settings_group(View3D.Ids.DATASELECTORS)
            .component_unique_id(component_id)
            .to_string()
        )

    def _color_scale_id(self, component_id: str) -> str:
        return (
            self.settings_group(View3D.Ids.COLORSCALE)
            .component_unique_id(component_id)
            .to_string()
        )

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self._view_id(VTKView3D.Ids.VIEW),
                "layers",
            ),
            Output(
                self._view_id(VTKView3D.Ids.VIEW),
                "bounds",
            ),
            Input(
                self._data_settings_id(DataSettings.Ids.PROPERTIES),
                "value",
            ),
            Input(
                self._data_settings_id(DataSettings.Ids.DATES),
                "value",
            ),
            Input(
                self._data_settings_id(DataSettings.Ids.REALIZATIONS),
                "value",
            ),
            Input(self.get_store_unique_id(ElementIds.IJK_CROP_STORE), "data"),
            Input(self._color_scale_id(ColorScale.Ids.COLORRANGEMANUAL), "value"),
            Input(self._color_scale_id(ColorScale.Ids.COLORMIN), "value"),
            Input(self._color_scale_id(ColorScale.Ids.COLORMAX), "value"),
            Input(
                self._color_scale_id(ColorScale.Ids.COLORMAP),
                "value",
            ),
            State(
                self._data_settings_id(DataSettings.Ids.STATIC_DYNAMIC),
                "value",
            ),
            State(
                self._view_id(VTKView3D.Ids.VIEW),
                "layers",
            ),
            State(
                self._view_id(VTKView3D.Ids.VIEW),
                "bounds",
            ),
        )
        # pylint: disable=too-many-arguments, too-many-locals
        def _set_geometry_and_scalar(
            prop: List[str],
            date: List[str],
            realizations: List[int],
            grid_range: List[List[int]],
            should_use_manual: List[str],
            manual_min: Optional[float],
            manual_max: Optional[float],
            colormap: str,
            proptype: str,
            layers: List[Dict],
            bounds: Optional[List[float]],
        ) -> Tuple[List[Dict], Optional[List]]:

            if PROPERTYTYPE(proptype) == PROPERTYTYPE.STATIC:
                property_spec = PropertySpec(prop_name=prop[0], prop_date=None)
            else:
                property_spec = PropertySpec(prop_name=prop[0], prop_date=date[0])

            realization = realizations[0]

            if not bounds or not layers[0]["bounds"]:
                geometrics = self.grid_provider.get_3dgrid(realization).get_geometrics(
                    allcells=True, return_dict=True
                )

                bounds = [
                    geometrics["xmin"],
                    geometrics["ymin"],
                    geometrics["xmax"],
                    geometrics["ymax"],
                ]
                layers[0]["bounds"] = [
                    geometrics["xmin"],
                    geometrics["ymin"],
                    -geometrics["zmax"],
                    geometrics["xmax"],
                    geometrics["ymax"],
                    -geometrics["zmin"],
                ]
            else:
                bounds = no_update
            provider_id = self.grid_provider.provider_id()

            cell_filter = CellFilter(
                i_min=grid_range[0][0],
                i_max=grid_range[0][1],
                j_min=grid_range[1][0],
                j_max=grid_range[1][1],
                k_min=grid_range[2][0],
                k_max=grid_range[2][1],
            )

            geometry_token = jwt.encode(
                {
                    "provider_id": provider_id,
                    "realization": realization,
                    "cell_filter": asdict(cell_filter),
                },
                "secret",
                algorithm="HS256",
            )
            geometry_and_property_token = jwt.encode(
                {
                    "provider_id": provider_id,
                    "realization": realization,
                    "cell_filter": asdict(cell_filter),
                    "property_spec": asdict(property_spec),
                },
                "secret",
                algorithm="HS256",
            )

            if should_use_manual and manual_min and manual_max:
                value_range = [manual_min, manual_max]
            else:
                value_range = None

            layers[1]["pointsUrl"] = f"/grid/points/{geometry_token}"
            layers[1]["polysUrl"] = f"/grid/polys/{geometry_token}"
            layers[1]["propertiesUrl"] = f"/grid/scalar/{geometry_and_property_token}"
            layers[1]["colorMapRange"] = value_range
            layers[1]["colorMapName"] = colormap
            return layers, bounds

        @callback(
            Output(
                self._color_scale_id(ColorScale.Ids.COLORMIN),
                "disabled",
            ),
            Output(
                self._color_scale_id(ColorScale.Ids.COLORMAX),
                "disabled",
            ),
            Input(
                self._color_scale_id(ColorScale.Ids.COLORRANGEMANUAL),
                "value",
            ),
        )
        def _toggle_manual_color(enabled: List[str]) -> Tuple[bool, bool]:

            if enabled:
                return False, False
            return True, True

        @callback(
            Output(self._view_id(VTKView3D.Ids.INFOBOX), "children"),
            Input(
                self._data_settings_id(DataSettings.Ids.PROPERTIES),
                "value",
            ),
            Input(
                self._data_settings_id(DataSettings.Ids.DATES),
                "value",
            ),
            Input(
                self._data_settings_id(DataSettings.Ids.REALIZATIONS),
                "value",
            ),
            Input(
                self._view_id(VTKView3D.Ids.VIEW),
                "layers",
            ),
            Input(self.get_store_unique_id(ElementIds.IJK_CROP_STORE), "data"),
            State(
                self._data_settings_id(DataSettings.Ids.STATIC_DYNAMIC),
                "value",
            ),
        )
        def update_infobox(
            properties: List[str],
            dates: List[str],
            realizations: List[int],
            layers: List[dict],
            grid_range: List[List[int]],
            proptype: str,
        ) -> list:
            """Updates the information box with information on the visualized data."""
            if PROPERTYTYPE(proptype) == PROPERTYTYPE.STATIC:
                property_spec = PropertySpec(prop_name=properties[0], prop_date=None)
            else:
                property_spec = PropertySpec(
                    prop_name=properties[0], prop_date=dates[0]
                )

            realization = realizations[0]
            provider_id = self.grid_provider.provider_id()

            cell_filter = CellFilter(
                i_min=grid_range[0][0],
                i_max=grid_range[0][1],
                j_min=grid_range[1][0],
                j_max=grid_range[1][1],
                k_min=grid_range[2][0],
                k_max=grid_range[2][1],
            )
            scalars = self.grid_viz_service.get_mapped_property_values(
                provider_id=provider_id,
                realization=realization,
                cell_filter=cell_filter,
                property_spec=property_spec,
            )
            if scalars is not None:
                actual_value_range = [
                    np.nanmin(scalars.value_arr),
                    np.nanmax(scalars.value_arr),
                ]
            else:
                actual_value_range = [0, 0]
            color_range = layers[1].get("colorMapRange")
            if color_range is None:
                color_range = actual_value_range

            return [
                html.Div([html.B("Property: "), html.Label(properties[0])]),
                html.Div(
                    [html.B("Date: "), html.Label(dates[0] if dates else "No date")]
                ),
                html.Div([html.B("Realization: "), html.Label(realization)]),
                html.Div(
                    [
                        html.B("Color range: "),
                        html.Label(f"{color_range[0]:.2f} - {color_range[1]:.2f}"),
                    ]
                ),
                html.Div(
                    [
                        html.B("Actual value range: "),
                        html.Label(
                            f"{actual_value_range[0]:.2f} - {actual_value_range[1]:.2f}"
                        ),
                    ]
                ),
            ]



from ._business_logic import SwatinitQcDataModel


def qc_flag_description() -> str:
    return """
- **PC_SCALED**  - Capillary pressure have been scaled and SWATINIT was accepted.
- **FINE_EQUIL**  - If item 9 in EQUIL is nonzero then initialization happens in a vertically
    refined model. Capillary pressure is still scaled, but water might be added or lost.
- **SWL_TRUNC**  - If SWL is larger than SWATINIT, SWAT will be reset to SWL. Extra water is
    added and hydrocarbons are lost.
- **SWATINIT_1**  - When SWATINIT is 1 above the contact, Eclipse will ignore SWATINIT and not
    touch the capillary pressure function which typically results in extra hydrocarbons.
    This could be ok as long as the porosities and/or permeabilities of these cells are small.
    If SWU is included, cells where SWATINIT is equal or larger than SWU will
    also be flagged as SWATINIT_1
- **HC_BELOW_FWL**  - If SWATINIT is less than 1 below the contact provided in EQUIL, Eclipse will
    ignore it and not scale the capillary pressure function. SWAT will be 1, unless a capillary
    pressure function with negative values is in SWOF/SWFN. This results in the loss of HC volumes.
- **PPCWMAX**  - If an upper limit of how much capillary pressure scaling is allowed is set, water will be
    lost if this limit is hit.
- **WATER**  - SWATINIT was 1 in the water zone, and SWAT is set to 1.

> **Consult the [check_swatinit](https://fmu-docs.equinor.com/docs/subscript/scripts/check_swatinit.html) 
  documentation for more detailed descriptions**

"""


def check_swatinit_description() -> str:
    return """
This plugin is used to visualize the output from **check_swatinit** which is a **QC tool for Water Initialization in Eclipse runs
when the SWATINIT keyword has been used**. It is used to quantify how much the volume changes from **SWATINIT** to **SWAT** at time
zero in the dynamical model, and help understand why it changes.

When the keyword SWATINIT has been used as water initialization option in Eclipse, capillary pressure scaling on a cell-by-cell basis will
occur in order to match SWATINIT from the geomodel. 
This process has some exceptions which can cause volume changes from SWATINIT to SWAT at time zero.
Each cell in the dynamical model has been flagged according to what has happened during initialization, and information is stored
in the **QC_FLAG** column.

> Check the maximum capillary pressure pr SATNUM in each EQLNUM to ensure extreme values were not necessary

[check_swatinit documentation](https://fmu-docs.equinor.com/docs/subscript/scripts/check_swatinit.html) 
"""


# pylint: disable=anomalous-backslash-in-string
def pc_columns_description() -> str:
    return f"""
> **Column descriptions**
> - **PCOW_MAX**  - Maximum capillary pressure from the input SWOF/SWFN tables
> - **PC_SCALING**  - Maximum capillary pressure scaling applied
> - **PPCW**  - Maximum capillary pressure after scaling
> - **{SwatinitQcDataModel.COLNAME_THRESHOLD}**  - Column showing how many percent of the pc-scaled dataset that match the user-selected threshold

*PPCW = PCOW_MAX \* PC_SCALING*

A threshold for the maximum capillary scaling can be set in the menu.
The table will show how many percent of the dataset that exceeds this value, and cells above the threshold will be shown in the map ➡️
"""



from typing import Dict, List, Union

import webviz_core_components as wcc
from dash import Input, Output, callback
from dash.development.base_component import Component
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .._plugin_ids import PluginIds
from ..utils import make_dataframes as makedf
from ..utils import make_figures as makefigs
from ._view_functions import _get_well_names_combined


class PlotSettingsHeatmap(SettingsGroupABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        COLOR_RANGE_SCALING = "color-range-scaling"
        FIG_LAYOUT_HEIGHT = "fig-layout-height"
        PLOT_TYPE = "plot-type"
        SHOW_WELLS_LARGEST_MISFIT = "show-wells-largest-misfit"

    def __init__(self) -> None:
        super().__init__("Plot Settings")

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Show wells with largest misfit",
                id=self.register_component_unique_id(
                    PlotSettingsHeatmap.Ids.SHOW_WELLS_LARGEST_MISFIT
                ),
                options=[
                    {"label": "Show all", "value": 0},
                    {"label": "2", "value": 2},
                    {"label": "4", "value": 4},
                    {"label": "6", "value": 6},
                    {"label": "8", "value": 8},
                    {"label": "10", "value": 10},
                    {"label": "12", "value": 12},
                    {"label": "15", "value": 15},
                    {"label": "20", "value": 20},
                    {"label": "25", "value": 25},
                ],
                value=0,
                multi=False,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Plot type",
                id=self.register_component_unique_id(PlotSettingsHeatmap.Ids.PLOT_TYPE),
                options=[
                    {"label": "Mean misfit", "value": "diffplot"},
                    {
                        "label": "Mean misfit relative (%)",
                        "value": "rel_diffplot",
                    },
                ],
                value="diffplot",
                multi=False,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Fig layout - height",
                id=self.register_component_unique_id(
                    PlotSettingsHeatmap.Ids.FIG_LAYOUT_HEIGHT
                ),
                options=[
                    {
                        "label": "Very small",
                        "value": 250,
                    },
                    {
                        "label": "Small",
                        "value": 350,
                    },
                    {
                        "label": "Medium",
                        "value": 450,
                    },
                    {
                        "label": "Large",
                        "value": 700,
                    },
                    {
                        "label": "Very large",
                        "value": 1000,
                    },
                ],
                value=450,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Color range scaling (relative to max)",
                id=self.register_component_unique_id(
                    PlotSettingsHeatmap.Ids.COLOR_RANGE_SCALING
                ),
                options=[
                    {"label": f"{x:.0%}", "value": x}
                    for x in [
                        0.1,
                        0.2,
                        0.3,
                        0.4,
                        0.5,
                        0.6,
                        0.7,
                        0.8,
                        0.9,
                        1.0,
                        1.5,
                        2.0,
                    ]
                ],
                style={"display": "block"},
                value=1.0,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
        ]


class ProdHeatmapView(ViewABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        PLOT_SETTINGS = "plot-settings"
        MAIN_COLUMN = "main-column"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        input_provider_set: EnsembleSummaryProviderSet,
        ens_vectors: Dict[str, List[str]],
        ens_realizations: Dict[str, List[int]],
        well_collections: Dict[str, List[str]],
    ) -> None:
        super().__init__("Well production heatmap")

        self.input_provider_set = input_provider_set
        self.ens_vectors = ens_vectors
        self.ens_realizations = ens_realizations
        self.well_collections = well_collections

        self.add_settings_group(
            PlotSettingsHeatmap(), ProdHeatmapView.Ids.PLOT_SETTINGS
        )
        self.main_column = self.add_column(ProdHeatmapView.Ids.MAIN_COLUMN)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.layout_element(ProdHeatmapView.Ids.MAIN_COLUMN)
                .get_unique_id()
                .to_string(),
                "children",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_ENSEMBLES), "data"
            ),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_DATES), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_PHASE), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_WELLS), "data"),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_WELL_COLLECTIONS),
                "data",
            ),
            Input(
                self.get_store_unique_id(
                    PluginIds.Stores.SELECTED_COMBINE_WELLS_COLLECTION
                ),
                "data",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_REALIZATIONS), "data"
            ),
            Input(
                self.settings_group(ProdHeatmapView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsHeatmap.Ids.SHOW_WELLS_LARGEST_MISFIT)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(ProdHeatmapView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsHeatmap.Ids.PLOT_TYPE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(ProdHeatmapView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsHeatmap.Ids.FIG_LAYOUT_HEIGHT)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(ProdHeatmapView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsHeatmap.Ids.COLOR_RANGE_SCALING)
                .to_string(),
                "value",
            ),
            # prevent_initial_call=True,
        )
        def _update_plots(
            ensemble_names: List[str],
            selector_dates: list,
            selector_phases: list,
            selector_well_names: list,
            selector_well_collection_names: list,
            selector_well_combine_type: str,
            selector_realizations: list,
            selector_filter_largest: int,
            selector_plot_type: str,
            selector_figheight: int,
            selector_scale_col_range: float,
        ) -> Union[str, List[Component]]:
            # pylint disable=too-many-arguments

            if not ensemble_names:
                return "No ensembles selected"

            well_names = _get_well_names_combined(
                self.well_collections,
                selector_well_collection_names,
                selector_well_names,
                selector_well_combine_type,
            )

            relative_diff = selector_plot_type == "rel_diffplot"
            dframe = makedf.get_df_diff_stat(
                makedf.get_df_diff(
                    makedf.get_df_smry(
                        self.input_provider_set,
                        ensemble_names,
                        self.ens_vectors,
                        self.ens_realizations,
                        selector_realizations,
                        well_names,
                        selector_phases,
                        selector_dates,
                    ),
                    relative_diff=relative_diff,
                )
            )

            figures = makefigs.heatmap_plot(
                dframe,
                selector_phases,
                vector_type="well",
                filter_largest=selector_filter_largest,
                figheight=selector_figheight,
                scale_col_range=selector_scale_col_range,
            )
            return figures



import datetime
from typing import List, Optional, Sequence

import pandas as pd

from webviz_subsurface._providers import (
    EnsembleSummaryProvider,
    Frequency,
    VectorMetadata,
)


class EnsembleSummaryProviderDummy(EnsembleSummaryProvider):
    """Class for EnsembleSummaryProvider mock implementations

    Creates dummy class implementing all methods in EnsembleSummaryProvider with
    NotImplementedError. Derived class is responsible to override necessary methods
    to obtain wanted functionality
    """

    #####################################################e
    #
    # Interface methods raise NotImplementedError
    #
    #####################################################

    def vector_names(self) -> List[str]:
        raise NotImplementedError("Method not implemented for mock!")

    def realizations(self) -> List[int]:
        raise NotImplementedError("Method not implemented for mock!")

    def vector_metadata(self, vector_name: str) -> Optional[VectorMetadata]:
        raise NotImplementedError("Method not implemented for mock!")

    def vector_names_filtered_by_value(
        self,
        exclude_all_values_zero: bool = False,
        exclude_constant_values: bool = False,
    ) -> List[str]:
        raise NotImplementedError("Method not implemented for mock!")

    def dates(
        self,
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> List[datetime.datetime]:
        raise NotImplementedError("Method not implemented for mock!")

    def supports_resampling(self) -> bool:
        raise NotImplementedError("Method not implemented for mock!")

    def get_vectors_df(
        self,
        vector_names: Sequence[str],
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        raise NotImplementedError("Method not implemented for mock!")

    def get_vectors_for_date_df(
        self,
        date: datetime.datetime,
        vector_names: Sequence[str],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        raise NotImplementedError("Method not implemented for mock!")



from typing import Any, List, Tuple

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from webviz_config import WebvizConfigTheme

from webviz_subsurface._figures import create_figure


# pylint: disable=too-many-public-methods
class PropertyStatisticsModel:
    """Class to process and visualize ensemble property statistics data"""

    REQUIRED_COLUMNS = [
        "ENSEMBLE",
        "REAL",
        "SOURCE",
        "ID",
        "PROPERTY",
        "Avg",
        "Avg_Weighted",
        "Max",
        "Min",
        "P10",
        "P90",
        "Stddev",
    ]
    REQUIRED_SELECTORS = ["ZONE"]
    SKIPPED_COLUMNS = ["Count"]

    def __init__(self, dataframe: pd.DataFrame, theme: WebvizConfigTheme) -> None:
        self._dataframe = dataframe
        self._dataframe["REAL"] = self._dataframe["REAL"].astype(int)
        self._prepare_and_validate_data()
        self._dataframe["label"] = self._dataframe.agg(
            lambda x: " | ".join(
                [f"{x[sel]}" for sel in ["PROPERTY"] + self.selectors]
            ),
            axis=1,
        )
        self.theme = theme
        self.colorway = self.theme.plotly_theme.get("layout", {}).get("colorway", None)
        self._statframe = self.aggregate_ensemble_data()

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._dataframe

    def _prepare_and_validate_data(self) -> None:
        for column in self.REQUIRED_COLUMNS + self.REQUIRED_SELECTORS:
            if column not in self.dataframe.columns:
                raise KeyError(
                    f"{column} column is missing from property statistics data"
                )
        self._dataframe = self._dataframe.drop(
            self.SKIPPED_COLUMNS, axis=1, errors="ignore"
        )
        sources = self.dataframe["SOURCE"].unique()
        if len(sources) > 1:
            raise ValueError(
                f"Property statistics data has multiple sources: {sources}. "
                "Only one source is supported"
            )
        ids = self.dataframe["ID"].unique()
        if len(sources) > 1:
            raise ValueError(
                f"Property statistics data has multiple ids: {ids}. "
                "Only one id is supported"
            )

    @property
    def statframe(self) -> pd.DataFrame:
        return self._statframe

    @property
    def selectors(self) -> List[str]:
        """List of user set selectors"""
        return [
            x
            for x in self.dataframe.columns
            if x not in self.REQUIRED_COLUMNS + ["label"]
        ]

    def selector_values(self, selector: str) -> List[Any]:
        return list(self.dataframe[selector].unique())

    def selectors_has_value(self, value: str = "Total") -> bool:
        for selector in self.selectors:
            if value not in self.selector_values(selector):
                return False
        return True

    @property
    def properties(self) -> List[str]:
        return list(self.dataframe["PROPERTY"].unique())

    @property
    def zones(self) -> List[str]:
        return list(self.dataframe["ZONE"].unique())

    @property
    def sources(self) -> List[str]:
        return list(self.dataframe["SOURCE"].unique())

    @property
    def ensembles(self) -> List[str]:
        return list(self.dataframe["ENSEMBLE"].unique())

    def aggregate_ensemble_data(self) -> pd.DataFrame:

        return (
            self.dataframe.drop(columns=["ID", "REAL"], errors="ignore")
            .groupby(["ENSEMBLE", "label", "PROPERTY", "SOURCE"] + self.selectors)[
                "Avg"
            ]
            .agg(
                [
                    ("Avg_Avg", "mean"),
                    ("Avg_P10", lambda x: np.percentile(x, 10)),
                    ("Avg_P90", lambda x: np.percentile(x, 90)),
                    ("Avg_Stddev", "std"),
                ]
            )
            .reset_index()
        )

    def get_labels(self, drop_constants: bool = True) -> List[str]:
        if drop_constants:
            return self.statframe[self.statframe["Avg_Stddev"] > 0]["label"].unique()
        return self.statframe["label"].unique()

    def get_real_order(
        self, ensemble: str, series: str, statistic: str = "Avg"
    ) -> pd.DataFrame:
        df = self.dataframe[self.dataframe["ENSEMBLE"] == ensemble]
        df = df[df["label"] == series]
        return df.sort_values(by=statistic)[[statistic, "REAL"]]

    def filter_on_label(
        self, ensemble: str, label: str, statistic: str = "Avg"
    ) -> pd.Series:
        return self.dataframe[
            (self.dataframe["ENSEMBLE"] == ensemble)
            & (self.dataframe["label"] == label)
        ][statistic]

    @staticmethod
    def filter_dataframe(
        dframe: pd.DataFrame, columns: list, column_values: list
    ) -> pd.DataFrame:

        if not isinstance(columns, list):
            columns = [columns]
        for filt, col in zip(column_values, columns):
            if isinstance(filt, list):
                dframe = dframe.loc[dframe[col].isin(filt)]
            else:
                dframe = dframe.loc[dframe[col] == filt]
        return dframe

    def filter_reals_on_label_range(
        self, ensemble: str, label: str, min_max: list, statistic: str = "Avg"
    ) -> pd.Series:
        return self.dataframe[
            (self.dataframe["ENSEMBLE"] == ensemble)
            & (self.dataframe["label"] == label)
            & (self.dataframe[statistic].between(min_max[0], min_max[1]))
        ]["REAL"]

    def get_ensemble_properties(
        self,
        ensemble: str,
        selector_values: List[Any],
        statistic: str = "Avg",
    ) -> pd.DataFrame:
        df = self.dataframe[self.dataframe["ENSEMBLE"] == ensemble]

        if selector_values is not None:
            df = self.filter_dataframe(df, self.selectors, selector_values)
        df = df[["REAL", statistic, "label"]]
        if df.empty:
            return df
        return (
            df.pivot_table(columns=["label"], values=statistic, index="REAL")
            .rename_axis(None, axis=1)
            .reset_index()
        )

    def delta_statistics(
        self,
        prop: str,
        ensemble: str,
        delta_ensemble: str,
        selector_values: List[Any],
        statistic: str = "Avg",
        aggregation: str = "Avg",
    ) -> pd.DataFrame:

        df = self.statframe.copy()

        if selector_values is not None:
            df = self.filter_dataframe(df, self.selectors, selector_values)
        df = df[df["PROPERTY"] == prop]
        df["label"] = df.agg(
            lambda x: " | ".join([f"{x[sel]}" for sel in self.selectors]), axis=1
        )

        # Drop non-numerical columns before aggregations
        df = df.drop(columns=["SOURCE", "PROPERTY"] + self.selectors)

        df = df[df["ENSEMBLE"].isin([ensemble, delta_ensemble])]
        df = df.pivot_table(columns=["ENSEMBLE"], index="label").reset_index()

        df["diff"] = (
            df[f"{statistic}_{aggregation}"][ensemble]
            - df[f"{statistic}_{aggregation}"][delta_ensemble]
        )
        if aggregation == "Avg":
            df["diff"] = abs(df["diff"])
        df = df.sort_values(by=["diff"])
        return df

    def make_delta_table(
        self,
        prop: str,
        ensemble: str,
        delta_ensemble: str,
        selector_values: List[Any],
        statistic: str = "Avg",
        aggregation: str = "Avg",
    ) -> Tuple[List[Any], List[Any]]:
        df = self.delta_statistics(
            prop=prop,
            ensemble=ensemble,
            delta_ensemble=delta_ensemble,
            selector_values=selector_values,
            statistic=statistic,
            aggregation=aggregation,
        )
        return self.make_table(df)

    def make_delta_bars(
        self,
        prop: str,
        ensemble: str,
        delta_ensemble: str,
        selector_values: List[Any],
        statistic: str = "Avg",
        aggregation: str = "Avg",
    ) -> go.Figure:
        df = self.delta_statistics(
            prop=prop,
            ensemble=ensemble,
            delta_ensemble=delta_ensemble,
            selector_values=selector_values,
            statistic=statistic,
            aggregation=aggregation,
        )
        fig = go.Figure(
            data=[
                {
                    "median": df["Avg_Avg"][ensemble],
                    "q1": df["Avg_P10"][ensemble],
                    "q3": df["Avg_P90"][ensemble],
                    "y": df["label"],
                    "type": "box",
                    "name": ensemble,
                    "hovertemplate": "test",
                    "hovertext": "none",
                },
                {
                    "median": df["Avg_Avg"][delta_ensemble],
                    "q1": df["Avg_P10"][delta_ensemble],
                    "q3": df["Avg_P90"][delta_ensemble],
                    "y": df["label"],
                    "type": "box",
                    "name": delta_ensemble,
                    "hovertemplate": "test",
                    "hovertext": "none",
                },
            ]
        )

        # Keep zoom level on click
        fig.update_layout(
            uirevision=f"{ensemble}{delta_ensemble}{prop}{str(df.size)}{aggregation}",
            boxmode="group",
            margin={"t": 20},
            yaxis_type="category",
        )
        fig = fig.to_dict()
        fig["layout"] = self.theme.create_themed_layout(fig["layout"])
        return fig

    @staticmethod
    def make_table(df: pd.DataFrame) -> Tuple[List[Any], List[Any]]:
        df.columns = df.columns.map("|".join)
        columns = [
            {
                "id": col,
                "name": [col.split("|")[0], col.split("|")[1]],
                "type": "numeric",
                "format": {"specifier": ".3~r"},
            }
            for col in df.columns
        ]
        return columns, df.iloc[::-1].to_dict("records")

    def make_statistics_table(
        self,
        ensembles: list,
        prop: str,
        selector_values: List[Any],
    ) -> Tuple[List[Any], List[Any]]:
        df = self.statframe.copy()
        df = df[df["PROPERTY"] == prop]
        if selector_values is not None:
            df = self.filter_dataframe(df, self.selectors, selector_values)

        # Drop non-numerical columns before aggregations
        df = df.drop(columns=["SOURCE", "PROPERTY"] + self.selectors)
        df = df[df["ENSEMBLE"].isin(ensembles)]
        df = df.pivot_table(columns=["ENSEMBLE"], index="label").reset_index()
        return self.make_table(df)

    def make_grouped_plot(
        self,
        ensembles: list,
        prop: str,
        selector_values: List[Any],
        statistic: str = "Avg",
        plot_type: str = "histogram",
        match_axis: bool = False,
    ) -> go.Figure:
        sel_length = 1
        for selector in selector_values:
            sel_length *= len(selector)

        df = self.dataframe.copy()
        df = df[df["PROPERTY"] == prop]
        if selector_values is not None:
            df = self.filter_dataframe(df, self.selectors, selector_values)

        df = df[df["ENSEMBLE"].isin(ensembles)]

        fig = create_figure(
            plot_type=plot_type if plot_type != "scatter_ensemble" else "scatter",
            data_frame=df,
            x="REAL" if plot_type != "distribution" else statistic,
            y=statistic if plot_type != "distribution" else None,
            facet_col="label" if plot_type != "scatter_ensemble" else "ENSEMBLE",
            color="ENSEMBLE" if plot_type != "scatter_ensemble" else "label",
            color_discrete_sequence=self.colorway,
        )
        if not match_axis:
            fig.update_xaxes(matches=None).update_yaxes(matches=None)

        fig = fig.to_dict()
        fig["layout"] = self.theme.create_themed_layout(fig["layout"])
        return fig

    def get_real_and_value_df(
        self,
        ensemble: str,
        series: str,
        normalize: bool = False,
        statistic: str = "Avg",
    ) -> pd.DataFrame:
        """
        Return dataframe with label and values for selected label for an ensemble.
        A column with normalized values can be added.
        """
        df = self.dataframe[self.dataframe["ENSEMBLE"] == ensemble]
        df = df[df["label"] == series]
        if normalize:
            df["VALUE_NORM"] = (df[statistic] - df[statistic].min()) / (
                df[statistic].max() - df[statistic].min()
            )

        df.rename(columns={statistic: "VALUE"}, inplace=True)

        return df[["label", "REAL", "VALUE", "VALUE_NORM"]].reset_index(drop=True)



import warnings
from typing import Dict, List, Optional

import dash_daq
import numpy as np
import pandas as pd
import webviz_core_components as wcc
from dash import Dash, Input, Output, State, dash_table, dcc, html
from dash.exceptions import PreventUpdate


class ColorPicker:
    """Component that can be added to a plugin to update colors interactively.
    The component is rendered as a table from a provided dataframe.
    'COLOR' is a required column and is rendered as a clickable row that displays
    a color picker for that row to change the color. The current colors can be
    retrieved as a list from a dcc.Store component"""

    def __init__(self, app: Dash, uuid: str, dframe: pd.DataFrame) -> None:
        """
        * **`app`:** The Dash app instance.
        * **`uuid`:** Unique id (use the plugin id).
        * **`dframe`:** Dataframe, where each row will be a row in a visualized table. \
            'COLOR' column with hexadecimal color strings is required. All other \
            colors must be validatable as categorical columns.
        """
        self._dframe = dframe
        self._check_colors()
        self._set_categorical()
        self._uuid = uuid
        self._set_callbacks(app)

    def _check_colors(self) -> None:
        """Check if color format is correct"""
        if "COLOR" not in self._dframe.columns:
            raise KeyError("Dataframe is missing 'COLOR' column")
        for hex_value in self._dframe["COLOR"].unique():
            if not hex_value.startswith("#"):
                raise ValueError(
                    "Incorrent color string found. "
                    f"{hex_value} is not a hex color string."
                )

    def _set_categorical(self) -> None:
        """Force all columns to be categorical"""
        for column in self._dframe.columns:
            try:
                self._dframe[column] = pd.Categorical(self._dframe[column])
            except ValueError as exc:
                raise ValueError(f"Cannot use {column} for ColorPicker") from exc

    @property
    def _columns(self) -> List[Dict[str, str]]:
        """Generate column attribute for the Dash Datatable"""
        columns = [
            {"id": col, "name": col, "selectable": False}
            for col in self._dframe.columns
            if col != "COLOR"
        ]
        columns.append({"id": "COLOR", "name": "Color", "selectable": True})
        return columns

    def get_color(self, color_list: List, filter_query: Dict[str, str]) -> str:
        """Helper function to use in a callback to get a single color.
        Give in the list of active colors, and a dictionary of {column:name:value}
        for each column in the color table"""
        filter_mask = [
            (self._dframe[column] == value) for column, value in filter_query.items()
        ]
        mask = np.array(filter_mask).all(axis=0)
        df = self._dframe.loc[mask]
        if len(df["COLOR"].unique()) < 1:
            warnings.warn("No color found for filter!")
            return "#ffffff"
        if len(df["COLOR"].unique()) > 1:
            warnings.warn("Multiple colors found for filter. " "Return first color.")
        return color_list[df.index[0]]

    @property
    def layout(self) -> html.Div:
        """Returns a table with dataframe columns and clickable color column
        Add this to the layout of the plugin"""
        return html.Div(
            style={
                "fontSize": "0.8em",
                "height": "600px",
                "fontColor": "black",
            },
            children=[
                dcc.Store(
                    id=self.color_store_id,
                    data=list(self._dframe["COLOR"].values),
                    storage_type="session",
                ),
                wcc.FlexBox(
                    style={"display": "flex"},
                    children=[
                        html.Div(
                            style={"flex": 2, "overflowY": "scroll", "height": "550px"},
                            children=dash_table.DataTable(
                                id={"id": self._uuid, "element": "table"},
                                fixed_rows={"headers": True},
                                columns=self._columns,
                                style_header={
                                    "opacity": 0.5,
                                },
                                data=self._dframe.to_dict("records"),
                                style_data_conditional=self.data_style_in_table,
                            ),
                        ),
                        html.Div(
                            id={"id": self._uuid, "element": "pickerwrapper"},
                            style={"flex": 1},
                            children=[
                                html.Label(
                                    "Click on a table row to get a color picker",
                                    style={"padding": "15px"},
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )

    @property
    def data_style_in_table(self) -> List[Dict[str, str]]:
        style_data = [
            {
                "if": {"row_index": idx, "column_id": "COLOR"},
                "background-color": self._dframe.iloc[idx]["COLOR"],
                "color": self._dframe.iloc[idx]["COLOR"],
            }
            for idx in range(0, self._dframe.shape[0])
        ]
        style_data.extend(
            [
                {
                    "if": {"state": "active"},
                    "backgroundColor": "white",
                    "border": "white",
                },
            ]
        )
        return style_data

    @property
    def color_store_id(self) -> Input:
        """Dom id for the current colors. Use the 'data' attribute in a callback
        to get the list of current colors"""
        return {"id": self._uuid, "element": "store"}

    def _set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(
                {"id": self._uuid, "element": "pickerwrapper"},
                "children",
            ),
            Input(
                {"id": self._uuid, "element": "table"},
                "active_cell",
            ),
            State({"id": self._uuid, "element": "store"}, "data"),
        )
        def _show_color_picker(
            cell: Optional[Dict], current_color_store: List[str]
        ) -> dash_daq.ColorPicker:
            """Render the colorpicker"""
            if not cell:
                raise PreventUpdate
            row_no = cell["row"]
            return dash_daq.ColorPicker(  # pylint: disable=not-callable
                {"id": self._uuid, "element": "picker"},
                label=f"Color for {[col for col in self._dframe.iloc[row_no] if col != 'COLOR']}",
                value=dict(hex=current_color_store[row_no]),
            )

        @app.callback(
            Output(
                {"id": self._uuid, "element": "table"},
                "style_data_conditional",
            ),
            Input({"id": self._uuid, "element": "store"}, "data"),
        )
        def _set_visual_color_in_table(current_color_store: List[str]) -> List:
            """Update the table color cell style"""
            style_data = [
                {
                    "if": {"row_index": idx, "column_id": "COLOR"},
                    "background-color": color,
                    "color": color,
                }
                for idx, color in enumerate(current_color_store)
            ]
            style_data.extend(
                [
                    {
                        "if": {"state": "active"},
                        "backgroundColor": "white",
                    },
                ]
            )
            return style_data

        @app.callback(
            Output({"id": self._uuid, "element": "store"}, "data"),
            Input(
                {"id": self._uuid, "element": "picker"},
                "value",
            ),
            State(
                {"id": self._uuid, "element": "table"},
                "active_cell",
            ),
            State({"id": self._uuid, "element": "store"}, "data"),
        )
        def _set_color(
            color: Optional[Dict],
            cell: Optional[Dict],
            current_color_store: List[str],
        ) -> List[str]:
            """Update list of stored colors"""
            if not cell or not color:
                raise PreventUpdate
            row_no = cell["row"]
            current_color = current_color_store[row_no]
            if current_color == color["hex"]:
                raise PreventUpdate
            current_color_store[row_no] = color["hex"]
            return current_color_store



from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.fanchart_plotting import (
    FanchartData,
    FreeLineData,
    LowHighData,
    MinMaxData,
    get_fanchart_traces,
)
from webviz_subsurface._utils.statistics_plotting import (
    LineData,
    StatisticsData,
    create_statistics_traces,
)

from .._types import FanchartOptions, StatisticsOptions
from .._utils.from_timeseries_cumulatives import is_per_interval_or_per_day_vector


def create_vector_observation_traces(
    vector_observations: dict,
    color: str = "black",
    legend_group: Optional[str] = None,
    show_legend: bool = False,
) -> List[dict]:
    """Create list of observations traces from vector observations

    `Input:`
    * vector_observations: dict - Dictionary with observation data for a vector
    * color: str - Color of observation traces for vector
    * legend_group: Optional[str] - Overwrite default legend group. Name of legend group, added as
    legend group and name if provided.
    * show_legend: bool - Show legend status for traces


    `Return:`
    List of marker traces for each observation for vector
    """
    observation_traces: List[dict] = []

    _name = "Observation" if legend_group is None else "Observation: " + legend_group
    _legend_group = "Observation" if legend_group is None else legend_group

    for observation in vector_observations.get("observations", []):
        hovertext = observation.get("comment")
        hovertemplate = (
            "(%{x}, %{y})<br>" + hovertext if hovertext else "(%{x}, %{y})<br>"
        )
        observation_traces.append(
            {
                "name": _name,
                "legendgroup": _legend_group,
                "x": [observation.get("date"), []],
                "y": [observation.get("value"), []],
                "marker": {"color": color},
                "hovertemplate": hovertemplate,
                "showlegend": show_legend,
                "error_y": {
                    "type": "data",
                    "array": [observation.get("error"), []],
                    "visible": True,
                },
            }
        )
    return observation_traces


def create_vector_realization_traces(
    vector_df: pd.DataFrame,
    ensemble: str,
    color: str,
    legend_group: str,
    line_shape: str,
    hovertemplate: str,
    show_legend: bool = False,
    legendrank: Optional[int] = None,
) -> List[dict]:
    """Renders line trace for each realization, includes history line if present

    `Input:`
    * vector_df: pd.DataFrame - Dataframe with vector data with following columns:\n
    ["DATE", "REAL", vector]

    * ensemble: str - Name of ensemble
    * color: str - color for traces
    * legend_group: str - legend group owner
    * line_shape: str - specified line shape for trace
    * show_legend: bool - show legend when true, otherwise do not show
    * hovertemplate: str - template for hovering of data points in trace lines
    * legendrank: int - rank value for legend in figure
    """
    vector_names = list(set(vector_df.columns) ^ set(["DATE", "REAL"]))
    if len(vector_names) != 1:
        raise ValueError(
            f"Expected one vector column present in dataframe, got {len(vector_names)}!"
        )

    vector_name = vector_names[0]
    return [
        {
            "line": {"width": 1, "shape": line_shape, "color": color},
            "mode": "lines",
            "x": list(real_df["DATE"]),
            "y": list(real_df[vector_name]),
            "hovertemplate": f"{hovertemplate}Realization: {real}, Ensemble: {ensemble}",
            "name": legend_group,
            "legendgroup": legend_group,
            "legendrank": legendrank,
            "showlegend": real_no == 0 and show_legend,
        }
        for real_no, (real, real_df) in enumerate(vector_df.groupby("REAL"))
    ]


def validate_vector_statistics_df_columns(
    vector_statistics_df: pd.DataFrame,
) -> None:
    """Validate columns of vector statistics DataFrame

    Verify DataFrame columns: ["DATE", MEAN, MIN, MAX, P10, P90, P50]

    Raise value error if columns are not matching

    `Input:`
    * vector_statistics_df: pd.Dataframe - Dataframe with dates and vector statistics columns.
    """
    expected_columns = [
        "DATE",
        StatisticsOptions.MEAN,
        StatisticsOptions.MIN,
        StatisticsOptions.MAX,
        StatisticsOptions.P10,
        StatisticsOptions.P90,
        StatisticsOptions.P50,
    ]
    if list(vector_statistics_df.columns) != expected_columns:
        raise ValueError(
            f"Incorrect dataframe columns, expected {expected_columns}, got "
            f"{vector_statistics_df.columns}"
        )


# pylint: disable = too-many-arguments, too-many-locals
def create_vector_statistics_traces(
    vector_statistics_df: pd.DataFrame,
    statistics_options: List[StatisticsOptions],
    color: str,
    legend_group: str,
    line_shape: str,
    line_width: int = 2,
    hovertemplate: str = "(%{x}, %{y})<br>",
    show_legend: bool = False,
    legendrank: Optional[int] = None,
) -> List[Dict[str, Any]]:
    """Get statistical lines for provided vector statistics DataFrame.

    `Input:`
    * vector_statistics_df: pd.Dataframe - Dataframe with dates and statistics columns
    for specific vector:\n
        DataFrame columns: ["DATE", MEAN, MIN, MAX, P10, P90, P50]

    * statistics_options: List[StatisticsOptions] - List of statistic options to include
    * color: str - color for traces
    * legend_group: str - legend group owner
    * line_shape: str - specified line shape for traces
    * line_width: str - specified line width for traces
    * hovertemplate: str - template for hovering of data points in trace lines
    * show_legend: bool - show legend when true, otherwise do not show
    * legendrank: int - rank value for legend in figure
    """
    # Validate columns format
    validate_vector_statistics_df_columns(vector_statistics_df)

    low_data = (
        LineData(
            data=vector_statistics_df[StatisticsOptions.P90].values,
            name=StatisticsOptions.P90,
        )
        if StatisticsOptions.P90 in statistics_options
        else None
    )
    mid_data = (
        LineData(
            data=vector_statistics_df[StatisticsOptions.P50].values,
            name=StatisticsOptions.P50,
        )
        if StatisticsOptions.P50 in statistics_options
        else None
    )
    high_data = (
        LineData(
            data=vector_statistics_df[StatisticsOptions.P10].values,
            name=StatisticsOptions.P10,
        )
        if StatisticsOptions.P10 in statistics_options
        else None
    )
    mean_data = (
        LineData(
            data=vector_statistics_df[StatisticsOptions.MEAN].values,
            name=StatisticsOptions.MEAN,
        )
        if StatisticsOptions.MEAN in statistics_options
        else None
    )
    minimum = (
        vector_statistics_df[StatisticsOptions.MIN].values
        if StatisticsOptions.MIN in statistics_options
        else None
    )
    maximum = (
        vector_statistics_df[StatisticsOptions.MAX].values
        if StatisticsOptions.MAX in statistics_options
        else None
    )

    data = StatisticsData(
        samples=vector_statistics_df["DATE"].values,
        free_line=mean_data,
        minimum=minimum,
        maximum=maximum,
        low=low_data,
        mid=mid_data,
        high=high_data,
    )
    return create_statistics_traces(
        data=data,
        color=color,
        legend_group=legend_group,
        line_shape=line_shape,
        line_width=line_width,
        show_legend=show_legend,
        hovertemplate=hovertemplate,
        legendrank=legendrank,
    )


def create_vector_fanchart_traces(
    vector_statistics_df: pd.DataFrame,
    fanchart_options: List[FanchartOptions],
    hex_color: str,
    legend_group: str,
    line_shape: str,
    hovertemplate: str = "(%{x}, %{y})<br>",
    show_legend: bool = False,
    legendrank: Optional[int] = None,
) -> List[Dict[str, Any]]:
    """Get statistical fanchart traces for provided vector statistics DataFrame.

    `Input:`
    * vector_statistics_df: pd.Dataframe - Dataframe with dates and statistics columns
    for specific vector:\n
        DataFrame columns: ["DATE", MEAN, MIN, MAX, P10, P90, P50]

    * fanchart_options: List[FanchartOptions] - List of fanchart options to include
    * hex_color: str - Hex color for traces and fill
    * legend_group: str - legend group owner
    * line_shape: str - specified line shape for trace
    * hovertemplate: str - template for hovering of data points in trace lines
    * show_legend: bool - show legend when true, otherwise do not show
    * legendrank: int - rank value for legend in figure
    """
    # Validate columns format
    validate_vector_statistics_df_columns(vector_statistics_df)

    low_high_data = (
        LowHighData(
            low_data=vector_statistics_df[StatisticsOptions.P90].values,
            low_name="P90",
            high_data=vector_statistics_df[StatisticsOptions.P10].values,
            high_name="P10",
        )
        if FanchartOptions.P10_P90 in fanchart_options
        else None
    )
    minimum_maximum_data = (
        MinMaxData(
            minimum=vector_statistics_df[StatisticsOptions.MIN].values,
            maximum=vector_statistics_df[StatisticsOptions.MAX].values,
        )
        if FanchartOptions.MIN_MAX in fanchart_options
        else None
    )
    mean_data = (
        FreeLineData(
            "Mean",
            vector_statistics_df[StatisticsOptions.MEAN].values,
        )
        if FanchartOptions.MEAN in fanchart_options
        else None
    )

    data = FanchartData(
        samples=vector_statistics_df["DATE"].tolist(),
        low_high=low_high_data,
        minimum_maximum=minimum_maximum_data,
        free_line=mean_data,
    )
    return get_fanchart_traces(
        data=data,
        hex_color=hex_color,
        legend_group=legend_group,
        line_shape=line_shape,
        show_legend=show_legend,
        hovertemplate=hovertemplate,
        legendrank=legendrank,
    )


def create_history_vector_trace(
    samples: list,
    history_data: np.ndarray,
    line_shape: str,
    color: str = "black",
    vector_name: Optional[str] = None,
    show_legend: bool = False,
    legendrank: Optional[int] = None,
) -> dict:
    """Returns the history data as trace line

    `Input:`
    * samples: list - list of samples
    * history_data: np.ndarray - 1D np.array of history data
    * line_shape: str - specified line shape
    * color: str - line color
    * vector_name: Optional[str] - Name of vector, appended to hovertext if provided
    * show_legend: bool - show legend when true, otherwise do not show

    `Return:`
    Trace line for provided history data. Raise value error if number of samples
    and number of history data points does not match.
    """
    if len(samples) != len(history_data):
        raise ValueError("Number of samples unequal number of data points!")

    hovertext = "History" if vector_name is None else "History: " + vector_name

    return {
        "line": {"shape": line_shape, "color": color},
        "mode": "lines",
        "x": samples,
        "y": history_data,
        "hovertext": hovertext,
        "hoverinfo": "y+x+text",
        "name": "History",
        "showlegend": show_legend,
        "legendgroup": "History",
        "legendrank": legendrank,
    }


def render_hovertemplate(vector: str, sampling_frequency: Optional[Frequency]) -> str:
    """Based on render_hovertemplate(vector: str, interval: Optional[str]) in
    webviz_subsurface/_utils/simulation_timeseries.py

    Adjusted to use Frequency enum and handle "Raw" and "weekly" frequency.

    `Input:`
    * vector: str - name of vector
    * sampling_frequency: Optional[Frequency] - sampling frequency for hovering data info
    """
    if is_per_interval_or_per_day_vector(vector) and sampling_frequency:
        if sampling_frequency in [Frequency.DAILY, Frequency.WEEKLY]:
            return "(%{x|%b} %{x|%-d}, %{x|%Y}, %{y})<br>"
        if sampling_frequency == Frequency.MONTHLY:
            return "(%{x|%b} %{x|%Y}, %{y})<br>"
        if sampling_frequency == Frequency.QUARTERLY:
            return "(Q%{x|%q} %{x|%Y}, %{y})<br>"
        if sampling_frequency == Frequency.YEARLY:
            return "(%{x|%Y}, %{y})<br>"
        raise ValueError(f"Interval {sampling_frequency.value} is not supported.")
    return "(%{x}, %{y})<br>"  # Plotly's default behavior



from pathlib import Path
from typing import Any, Optional, Union

import pandas as pd
from webviz_config.webviz_store import webvizstore

from .fmu_input import load_ensemble_set

try:
    import ecl2df
except ImportError:
    pass

try:
    from pyscal import PyscalFactory
except ImportError:
    pass


@webvizstore
def load_satfunc(
    ensemble_paths: dict,
    ensemble_set_name: str = "EnsembleSet",
) -> pd.DataFrame:
    def ecl2df_satfunc(kwargs: Any) -> pd.DataFrame:
        return ecl2df.satfunc.df(kwargs["realization"].get_eclfiles())

    return load_ensemble_set(ensemble_paths, ensemble_set_name).apply(ecl2df_satfunc)


@webvizstore
def load_scal_recommendation(
    scalfile: Path, sheet_name: Optional[Union[str, int, list]] = None
) -> pd.DataFrame:
    return (
        PyscalFactory.create_scal_recommendation_list(
            PyscalFactory.load_relperm_df(str(scalfile), sheet_name)
        )
        .df()
        .replace(
            {
                "CASE": {
                    "(?i)low": "pess",
                    r"(?i)^pes.*": "pess",
                    "(?i)base": "base",
                    "(?i)high": "opt",
                    r"(?i)^opt.*": "opt",
                }
            },
            regex=True,
        )
    )



from typing import Any, List, Optional

import webviz_core_components as wcc
from dash import Input, Output, State, callback, dcc, html
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class Selections(SettingsGroupABC):
    class Ids(StrEnum):
        VFP_NAME = "vfp-name"
        METADATA_BUTTON = "metadata-button"
        METADATA_DIALOG = "metadata-dialog"

    def __init__(self, vfp_names: List[str]) -> None:
        super().__init__("Selections")
        self._vfp_names = vfp_names

    def layout(self) -> List[Any]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(Selections.Ids.VFP_NAME),
                label="VFP name",
                options=[{"label": vfp, "value": vfp} for vfp in self._vfp_names],
                clearable=False,
                value=self._vfp_names[0] if len(self._vfp_names) > 0 else None,
                persistence=True,
                persistence_type="session",
            ),
            html.Button(
                "Show VFP Metadata",
                style={
                    "width": "100%",
                    "background-color": "white",
                    "margin-top": "5px",
                },
                id=self.register_component_unique_id(Selections.Ids.METADATA_BUTTON),
            ),
            wcc.Dialog(
                title="VFP Table Metadata",
                id=self.register_component_unique_id(Selections.Ids.METADATA_DIALOG),
                max_width="md",
                open=False,
                children=dcc.Markdown("VFP Metadata not set."),
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(self.Ids.METADATA_DIALOG).to_string(), "open"
            ),
            Input(
                self.component_unique_id(self.Ids.METADATA_BUTTON).to_string(),
                "n_clicks",
            ),
            State(
                self.component_unique_id(self.Ids.METADATA_DIALOG).to_string(), "open"
            ),
        )
        @callback_typecheck
        def open_close_metadata_dialog(n_clicks: Optional[int], is_open: bool) -> bool:
            if n_clicks is not None:
                return not is_open
            raise PreventUpdate



import webviz_core_components as wcc
from dash import html

from webviz_subsurface._models import InplaceVolumesModel

from ..utils.table_and_figure_utils import create_figure_matrix


def tornado_main_layout(uuid: str) -> html.Div:
    return html.Div(
        children=[
            html.Div(id={"id": uuid, "page": "torn_multi"}, style={"display": "block"}),
            html.Div(
                id={"id": uuid, "page": "torn_bulk_inplace"}, style={"display": "none"}
            ),
        ]
    )


def tornado_plots_layout(figures: list, bottom_display: list) -> html.Div:
    matrix = create_figure_matrix(figures)
    max_height = 45 if bottom_display else 86

    return html.Div(
        children=[
            html.Div(
                children=[
                    wcc.FlexBox(
                        children=[
                            html.Div(
                                style={"flex": 1},
                                children=wcc.Graph(
                                    config={"displayModeBar": False},
                                    style={"height": f"{max_height/len(matrix)}vh"},
                                    figure=fig,
                                )
                                if fig is not None
                                else [],
                            )
                            for fig in row
                        ]
                    )
                    for row in matrix
                ],
            ),
            html.Div(
                bottom_display,
                style={
                    "height": "40vh",
                    "display": "block" if bottom_display else "none",
                    "margin-top": "20px",
                },
            ),
        ]
    )


def tornado_error_layout(message: str) -> wcc.Frame:
    return html.Div(message, style={"margin-top": "40px"})


def tornado_selections_layout(
    uuid: str, volumemodel: InplaceVolumesModel, tab: str
) -> html.Div:
    """Layout for selecting tornado data"""
    return html.Div(
        children=[
            page_buttons(uuid, volumemodel),
            wcc.Selectors(
                label="TORNADO CONTROLS",
                children=tornado_controls_layout(uuid, tab, volumemodel),
            ),
            wcc.Selectors(
                label="⚙️ SETTINGS",
                open_details=True,
                children=[
                    scale_selector(uuid, tab),
                    checkboxes_settings(uuid, tab),
                    labels_display(uuid, tab),
                    reference_selector(uuid, tab, volumemodel),
                ],
            ),
        ]
    )


def page_buttons(uuid: str, volumemodel: InplaceVolumesModel) -> html.Div:
    no_bulk = "BULK" not in volumemodel.responses
    return html.Div(
        style={"margin-bottom": "20px", "display": "none" if no_bulk else "block"},
        children=[
            button(uuid=uuid, title="Custom", page_id="torn_multi"),
            button(
                uuid=uuid,
                title="Bulk vs STOIIP/GIIP",
                page_id="torn_bulk_inplace",
            ),
        ],
    )


def button(uuid: str, title: str, page_id: str) -> html.Button:
    return html.Button(
        title,
        className="webviz-inplace-vol-btn",
        id={"id": uuid, "button": page_id},
    )


def tornado_controls_layout(
    uuid: str, tab: str, volumemodel: InplaceVolumesModel
) -> wcc.Selectors:
    sens_columns = ["REAL", "SENSNAME", "SENSCASE", "SENSTYPE", "SENSNAME_CASE"]
    return [
        wcc.Dropdown(
            label="Response",
            id={"id": uuid, "tab": tab, "selector": "Response"},
            clearable=False,
            options=[{"label": i, "value": i} for i in volumemodel.responses],
            value=volumemodel.responses[0],
        ),
        wcc.SelectWithLabel(
            label="Sensitivity filter",
            collapsible=True,
            open_details=False,
            id={"id": uuid, "tab": tab, "selector": "Sensitivities"},
            options=[{"label": i, "value": i} for i in volumemodel.sensitivities],
            value=volumemodel.sensitivities,
            size=min(15, len(volumemodel.sensitivities)),
        ),
        wcc.Dropdown(
            label="Subplots",
            id={"id": uuid, "tab": tab, "selector": "Subplots"},
            clearable=True,
            options=[
                {"label": i, "value": i}
                for i in [
                    x
                    for x in volumemodel.selectors
                    if x not in sens_columns and volumemodel.dataframe[x].nunique() > 1
                ]
            ],
        ),
        html.Div(
            style={"margin-top": "10px"},
            children=wcc.RadioItems(
                label="Visualization below tornado:",
                id={"id": uuid, "tab": tab, "selector": "bottom_viz"},
                options=[
                    {"label": "Table", "value": "table"},
                    {"label": "Realization plot", "value": "realplot"},
                    {"label": "None", "value": "none"},
                ],
                vertical=False,
                value="table",
            ),
        ),
    ]


def checkboxes_settings(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-top": "10px", "margin-bottom": "10px"},
        children=[
            wcc.Checklist(
                id={"id": uuid, "tab": tab, "selector": selector},
                options=[{"label": label, "value": "selected"}],
                value=["selected"] if selected else [],
            )
            for label, selector, selected in [
                ("Color by sensitivity", "color_by_sens", True),
                ("Shared subplot X axis", "Shared axis", False),
                ("Show realization points", "real_scatter", False),
                ("Show reference on tornado", "torn_ref", True),
                ("Remove sensitivities with no impact", "Remove no impact", True),
            ]
        ],
    )


def labels_display(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-bottom": "10px"},
        children=[
            wcc.RadioItems(
                label="Label options:",
                id={"id": uuid, "tab": tab, "selector": "labeloptions"},
                options=[
                    {"label": "detailed", "value": "detailed"},
                    {"label": "simple", "value": "simple"},
                    {"label": "hide", "value": "hide"},
                ],
                vertical=False,
                value="simple",
            ),
        ],
    )


def reference_selector(
    uuid: str, tab: str, volumemodel: InplaceVolumesModel
) -> wcc.Dropdown:
    return wcc.Dropdown(
        label="Reference:",
        id={"id": uuid, "tab": tab, "selector": "Reference"},
        options=[{"label": elm, "value": elm} for elm in volumemodel.sensitivities],
        value="rms_seed"
        if "rms_seed" in volumemodel.sensitivities
        else volumemodel.sensitivities[0],
        clearable=False,
    )


def scale_selector(uuid: str, tab: str) -> wcc.Dropdown:
    return wcc.Dropdown(
        label="Scale:",
        id={"id": uuid, "tab": tab, "selector": "Scale"},
        options=[
            {"label": "Relative value (%)", "value": "Percentage"},
            {"label": "Relative value", "value": "Absolute"},
            {"label": "True value", "value": "True"},
        ],
        value="Percentage",
        clearable=False,
    )



from typing import List, Optional

import webviz_core_components as wcc
from dash import html

from webviz_subsurface._models import InplaceVolumesModel


def filter_layout(
    uuid: str,
    tab: str,
    volumemodel: InplaceVolumesModel,
    open_details: bool = True,
    hide_selectors: Optional[list] = None,
) -> wcc.Selectors:
    """Layout for selecting intersection data"""
    return wcc.Selectors(
        label="FILTERS",
        open_details=open_details,
        children=[
            filter_dropdowns(
                uuid=uuid,
                tab=tab,
                volumemodel=volumemodel,
                hide_selectors=hide_selectors,
            ),
            realization_filters(uuid=uuid, tab=tab, volumemodel=volumemodel),
        ],
    )


def filter_dropdowns(
    uuid: str,
    volumemodel: InplaceVolumesModel,
    tab: str,
    hide_selectors: Optional[list] = None,
) -> html.Div:
    """Makes dropdowns for each selector"""
    dropdowns_layout: List[html.Div] = []
    hide_selectors = ["SENSNAME", "SENSTYPE", "SENSCASE"] + (
        hide_selectors if hide_selectors is not None else []
    )
    selectors = [
        x
        for x in volumemodel.selectors
        if x not in volumemodel.region_selectors + ["REAL"]
    ]
    for selector in selectors:
        dropdowns_layout.append(
            create_filter_select(
                selector,
                elements=list(volumemodel.dataframe[selector].unique()),
                filter_type="undef",
                uuid=uuid,
                tab=tab,
                hide=selector in hide_selectors,
            )
        )
    # Make region filters
    dropdowns_layout.append(
        html.Span("Region filters: ", style={"font-weight": "bold"})
    )
    if all(x in volumemodel.region_selectors for x in ["FIPNUM", "ZONE", "REGION"]):
        dropdowns_layout.append(fipnum_vs_zone_region_switch(uuid, tab))

    for selector in volumemodel.region_selectors:
        dropdowns_layout.append(
            create_filter_select(
                selector,
                elements=list(volumemodel.dataframe[selector].unique()),
                filter_type="region",
                uuid=uuid,
                tab=tab,
                hide=selector == "FIPNUM" and len(volumemodel.region_selectors) > 1,
            )
        )
    return html.Div(dropdowns_layout)


def create_filter_select(
    selector: str, elements: list, uuid: str, tab: str, filter_type: str, hide: bool
) -> html.Div:
    return html.Div(
        id={"id": uuid, "tab": tab, "wrapper": selector, "type": filter_type},
        style={"display": "inline" if len(elements) > 1 and not hide else "none"},
        children=wcc.SelectWithLabel(
            label=selector.lower().capitalize(),
            id={"id": uuid, "tab": tab, "selector": selector, "type": filter_type},
            options=[{"label": i, "value": i} for i in elements],
            value=elements,
            multi=True,
            size=min(15, len(elements)),
        ),
    )


def fipnum_vs_zone_region_switch(uuid: str, tab: str) -> wcc.RadioItems:
    return wcc.RadioItems(
        id={"id": uuid, "tab": tab, "element": "region-selector"},
        options=[
            {"label": "Region∕Zone", "value": "regzone"},
            {"label": "Fipnum", "value": "fipnum"},
        ],
        value="regzone",
        vertical=False,
    )


def realization_filters(
    uuid: str, tab: str, volumemodel: InplaceVolumesModel
) -> html.Div:
    reals = volumemodel.realizations
    return html.Div(
        style={"margin-top": "15px"},
        children=[
            html.Div(
                style={"display": "inline-flex"},
                children=[
                    html.Span(
                        "Realizations: ",
                        style={"font-weight": "bold"},
                    ),
                    html.Span(
                        id={"id": uuid, "tab": tab, "element": "real_text"},
                        style={"margin-left": "10px"},
                        children=f"{min(reals)}-{max(reals)}",
                    ),
                ],
            ),
            wcc.RadioItems(
                id={"id": uuid, "tab": tab, "element": "real-selector-option"},
                options=[
                    {"label": "Range", "value": "range"},
                    {"label": "Select", "value": "select"},
                ],
                value="range",
                vertical=False,
            ),
            wcc.RangeSlider(
                wrapper_id={"id": uuid, "tab": tab, "element": "real-slider-wrapper"},
                id={
                    "id": uuid,
                    "tab": tab,
                    "component_type": "range",
                },
                value=[min(reals), max(reals)],
                min=min(reals),
                max=max(reals),
                marks={str(i): {"label": str(i)} for i in [min(reals), max(reals)]},
            ),
            html.Div(
                style={"display": "none"},
                children=wcc.Select(
                    id={"id": uuid, "tab": tab, "selector": "REAL", "type": "REAL"},
                    options=[{"label": i, "value": i} for i in reals],
                    value=reals,
                ),
            ),
        ],
    )



import pathlib
from typing import Any, Dict

import dash
import pytest
from _pytest.config.argparsing import Parser
from _pytest.fixtures import SubRequest
from webviz_config import WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.themes import default_theme
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WEBVIZ_INSTANCE_INFO, WebvizRunMode


def pytest_addoption(parser: Parser) -> None:
    parser.addoption(
        "--testdata-folder",
        type=pathlib.Path,
        default=pathlib.Path("webviz-subsurface-testdata"),
        help="Path to webviz-subsurface-testdata folder",
    )


@pytest.fixture(name="testdata_folder")
def testdata_folder_fixture(request: SubRequest) -> Any:
    return request.config.getoption("--testdata-folder")


@pytest.fixture()
def app() -> dash.Dash:
    dash_app = dash.Dash(__name__)

    WEBVIZ_INSTANCE_INFO.initialize(
        dash_app=dash_app,
        run_mode=WebvizRunMode.NON_PORTABLE,
        theme=default_theme,
        storage_folder=pathlib.Path(__file__).resolve().parent,
    )
    try:
        WEBVIZ_FACTORY_REGISTRY.initialize(None)
    except RuntimeError:
        pass

    dash_app.css.config.serve_locally = True
    dash_app.scripts.config.serve_locally = True
    dash_app.config.suppress_callback_exceptions = True
    CACHE.init_app(dash_app.server)
    yield dash_app


@pytest.fixture()
def shared_settings(testdata_folder: pathlib.Path) -> Dict:
    return {
        "HM_SETTINGS": WebvizSettings(
            theme=default_theme,
            shared_settings={
                "scratch_ensembles": {
                    "iter-0": f"{testdata_folder}/01_drogon_ahm/realization-*/iter-0",
                    "iter-3": f"{testdata_folder}/01_drogon_ahm/realization-*/iter-3",
                }
            },
        ),
        "HM_ENSEMBLES": ["iter-0", "iter-3"],
        "SENS_SETTINGS": WebvizSettings(
            theme=default_theme,
            shared_settings={
                "scratch_ensembles": {
                    "sens_run": f"{testdata_folder}/01_drogon_design/realization-*/iter-0",
                }
            },
        ),
        "SENS_ENSEMBLES": ["sens_run"],
    }



#!/usr/bin/env python
"""Batch conversion of UNSMRY files to Apache Arrow IPC file format
"""

import argparse
import glob
import logging
import os
import warnings
from pathlib import Path
from typing import List

import ecl2df

logger = logging.getLogger(__name__)


def _get_parser() -> argparse.ArgumentParser:
    """Setup parser for command line options"""

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.description = (
        "Batch conversion of UNSMRY files to Apache Arrow IPC file format.\n"
        "\n"
        "Required input is a path/paths which may be wildcarded, like:\n"
        "  smry2arrow_batch my_folder/realization-*/iter-0\n"
        "You may also define a specific file pattern for the UNSMRY file to read."
    )

    parser.add_argument(
        "enspath",
        type=Path,
        help="Path with wildcards giving file system location of the ensemble's realizations",
        nargs="*",
    )
    parser.add_argument(
        "--eclbase",
        type=Path,
        help='Eclipse base name, note that "" is required around paths with wildcards',
        default=Path("eclipse/model/*.UNSMRY"),
    )
    return parser


def _convert_single_smry_file(smry_filename: str, arrow_filename: str) -> None:
    """Read summary data for single realization from disk and write it out to .arrow
    file using ecl2df.
    """

    eclbase = (
        smry_filename.replace(".DATA", "").replace(".UNSMRY", "").replace(".SMSPEC", "")
    )

    eclfiles = ecl2df.EclFiles(eclbase)
    sum_df = ecl2df.summary.df(eclfiles)

    # Slight hack here, using ecl2df protected function to gain access to conversion routine
    # pylint: disable=protected-access
    sum_table = ecl2df.summary._df2pyarrow(sum_df)

    ecl2df.summary.write_dframe_stdout_file(sum_table, arrow_filename)


def _batch_convert_smry2arrow(
    ens_path: List[Path], ecl_base: Path, relative_output_dir: Path
) -> None:
    """Does batch conversion of UNSMRY files for all realizations within an ensemble."""

    for wildcarded_path in ens_path:
        globbed_real_dirs = sorted(glob.glob(str(wildcarded_path)))

        for real_dir in globbed_real_dirs:
            glob_expr = str(Path(real_dir) / ecl_base)
            globbed_smry_files = sorted(glob.glob(glob_expr))
            if globbed_smry_files:
                real_output_dir = Path(real_dir) / relative_output_dir
                real_output_dir.mkdir(parents=True, exist_ok=True)

                for smry_file in globbed_smry_files:
                    basename_without_ext = Path(Path(smry_file).name).stem
                    arrow_file = real_output_dir / (basename_without_ext + ".arrow")

                    logger.info(f"input(smry):   {smry_file}")
                    logger.info(f"output(arrow): {arrow_file}")
                    _convert_single_smry_file(smry_file, str(arrow_file))


def main() -> None:
    """Entry point from command line"""

    parser = _get_parser()
    args = parser.parse_args()

    logging.basicConfig(level=logging.WARNING)
    logger.setLevel(logging.INFO)

    enspath = args.enspath

    # Strip out any leading path separator from eclbase
    eclbase = str(args.eclbase).lstrip(os.path.sep)

    # If no extension has been given, we add .UNSMRY
    if not os.path.splitext(eclbase)[1]:
        eclbase += ".UNSMRY"

    # Output directory relative to each realization's root directory
    relative_output_dir = Path("share/results/unsmry")

    warnings.warn(
        "This script is a temporary solution for converting existing ensemble summary data "
        "for usage with webviz-subsurface, and will probably be removed in the future. "
        "New ensembles should be configured with an ert job for outputting .arrow files directly.",
        FutureWarning,
    )

    logger.info(f"enspath: {enspath}")
    logger.info(f"eclbase: {eclbase}")

    _batch_convert_smry2arrow(enspath, Path(eclbase), relative_output_dir)

    logger.info("done")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main()



from typing import Callable, Dict, List, Optional, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from ._utils import VfpDataModel
from ._views._vfp_view import VfpView


class VfpAnalysis(WebvizPluginABC):
    """Vizualizes VFP curves.

    ---

    * **`vfp_file_pattern`:** File pattern for where to search for vfp arrow files. The path
    should be relative to the runpath if ensemble and realization is given as input, if not
    the path needs to be absolute.
    * **`ensemble`:** Which ensemble in `shared_settings` to use.
    * **`realization`:** Which realization to pick from the ensemble.
    ---

    The plugin uses an `.arrow` representation of the VFP curves, which can be exported to disk by
    using the `ECL2CSV` forward model in ERT with subcommand `vfp`.

    So far, the plugin only vizualizes VFPPROD curves, but the plan is to extend it also to
    VFPINJ curves soon.

    """

    class Ids(StrEnum):
        VFP_VIEW = "vpf-view"

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        vfp_file_pattern: str = "share/results/tables/vfp/*.arrow",
        ensemble: Optional[str] = None,
        realization: Optional[int] = None,
    ) -> None:
        super().__init__(stretch=True)

        self._datamodel = VfpDataModel(
            webviz_settings=webviz_settings,
            vfp_file_pattern=vfp_file_pattern,
            ensemble=ensemble,
            realization=realization,
        )

        self.add_view(VfpView(self._datamodel), self.Ids.VFP_VIEW)

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return self._datamodel.webviz_store



from typing import List, Union

import pandas as pd
from dash import Input, Output, callback
from webviz_config import WebvizSettings
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewABC

from ...._utils.unique_theming import unique_colors
from .._plugin_ids import PluginIds
from ..view_elements import Graph
from ._view_functions import _get_fanchart_traces, calc_statistics, filter_df


class FanView(ViewABC):
    class Ids(StrEnum):
        FAN_CHART = "fan-chart"

    def __init__(self, bhp_df: pd.DataFrame, webviz_settings: WebvizSettings) -> None:
        super().__init__("Fan chart")

        self.bhp_df = bhp_df

        column = self.add_column()
        column.add_view_element(Graph(), FanView.Ids.FAN_CHART)
        self.theme = webviz_settings.theme

    @property
    def ensembles(self) -> List[str]:
        return list(self.bhp_df["ENSEMBLE"].unique())

    @property
    def ens_colors(self) -> dict:
        return unique_colors(self.ensembles, self.theme)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(FanView.Ids.FAN_CHART)
                .component_unique_id(Graph.Ids.GRAPH)
                .to_string(),
                "figure",
            ),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_ENSEMBLE), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_SORT_BY), "data"),
            Input(
                self.get_store_unique_id(
                    PluginIds.Stores.SELECTED_ASCENDING_DESCENDING
                ),
                "data",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_MAX_NUMBER_OF_WELLS),
                "data",
            ),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_WELLS), "data"),
        )
        def _update_plot(
            ensemble: str,
            sort_by: str,
            ascending: bool,
            n_wells: int,
            wells: Union[str, List[str]],
        ) -> dict:
            wells = wells if isinstance(wells, list) else [wells]
            df = filter_df(df=self.bhp_df, ensemble=ensemble, wells=wells)
            stat_df = (
                calc_statistics(df)
                .sort_values(sort_by, ascending=ascending)
                .iloc[0:n_wells, :]
            )
            fan_traces = _get_fanchart_traces(
                ens_stat_df=stat_df,
                color=self.ens_colors[ensemble],
                legend_group=ensemble,
            )
            layout = self.theme.create_themed_layout(
                {
                    "yaxis": {
                        "side": "left",
                        "title": "Bottom hole pressure",
                        "showgrid": True,
                    },
                    "yaxis2": {
                        "side": "right",
                        "overlaying": "y",
                        "title": "Count (data points)",
                        "showgrid": False,
                    },
                    "xaxis": {"showgrid": False},
                    "barmode": "group",
                    "legend": {"x": 1.05},
                }
            )
            return {"data": fan_traces, "layout": layout}



from typing import List

import pytest

from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.delta_ensemble_utils import (
    DeltaEnsemble,
    create_delta_ensemble_name,
    create_delta_ensemble_name_dict,
    create_delta_ensemble_names,
    create_delta_ensemble_provider_pair,
    is_delta_ensemble_providers_in_provider_set,
)

from ....mocks.ensemble_summary_provider_dummy import EnsembleSummaryProviderDummy

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************


class EnsembleSummaryProviderMock(EnsembleSummaryProviderDummy):
    """Ensemble summary provider mock for testing

    Note empty list returned in override methods, only to allow constructing
    EnsembleSummaryProviderSet objects!
    """

    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name

    def get_name(self) -> str:
        return self._name

    ########################################
    #
    # Override methods
    #
    ########################################
    def vector_names(self) -> List[str]:
        """Return empty list only to allow constructing EnsembleSummaryProviderSet object"""
        return []

    def realizations(self) -> List[int]:
        """Return empty list only to allow constructing EnsembleSummaryProviderSet object"""
        return []


# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************


def test_create_delta_ensemble_name() -> None:
    assert (
        create_delta_ensemble_name(
            DeltaEnsemble(ensemble_a="first name", ensemble_b="second name")
        )
        == "(first name)-(second name)"
    )
    assert (
        create_delta_ensemble_name(
            DeltaEnsemble(ensemble_a="first-Name", ensemble_b="second-Name")
        )
        == "(first-Name)-(second-Name)"
    )
    assert (
        create_delta_ensemble_name(
            DeltaEnsemble(ensemble_a="ens-0", ensemble_b="ens-3")
        )
        == "(ens-0)-(ens-3)"
    )


def test_create_delta_ensemble_names() -> None:
    first_delta_ensemble = DeltaEnsemble(
        ensemble_a="first name", ensemble_b="second name"
    )
    second_delta_ensemble = DeltaEnsemble(
        ensemble_a="first-Name", ensemble_b="second-Name"
    )
    third_delta_ensemble = DeltaEnsemble(ensemble_a="ens-0", ensemble_b="ens-3")

    assert create_delta_ensemble_names(
        [first_delta_ensemble, second_delta_ensemble, third_delta_ensemble]
    ) == ["(first name)-(second name)", "(first-Name)-(second-Name)", "(ens-0)-(ens-3)"]


def test_create_delta_ensemble_name_dict() -> None:
    first_delta_ensemble = DeltaEnsemble(
        ensemble_a="first name", ensemble_b="second name"
    )
    second_delta_ensemble = DeltaEnsemble(
        ensemble_a="first-Name", ensemble_b="second-Name"
    )
    third_delta_ensemble = DeltaEnsemble(ensemble_a="ens-0", ensemble_b="ens-3")

    expected_delta_ensemble_name_dict = {
        "(first name)-(second name)": first_delta_ensemble,
        "(first-Name)-(second-Name)": second_delta_ensemble,
        "(ens-0)-(ens-3)": third_delta_ensemble,
    }

    assert (
        create_delta_ensemble_name_dict(
            [first_delta_ensemble, second_delta_ensemble, third_delta_ensemble]
        )
        == expected_delta_ensemble_name_dict
    )


def test_is_delta_ensemble_providers_in_provider_set() -> None:
    provider_set = EnsembleSummaryProviderSet(
        {
            "First provider": EnsembleSummaryProviderMock("First mock"),
            "Second provider": EnsembleSummaryProviderMock("Second mock"),
        }
    )

    valid_delta_ensemble = DeltaEnsemble(
        ensemble_a="First provider", ensemble_b="Second provider"
    )
    invalid_delta_ensemble = DeltaEnsemble(
        ensemble_a="First provider", ensemble_b="Third provider"
    )

    assert is_delta_ensemble_providers_in_provider_set(
        valid_delta_ensemble, provider_set
    )
    assert not is_delta_ensemble_providers_in_provider_set(
        invalid_delta_ensemble, provider_set
    )


def test_create_delta_ensemble_provider_pair() -> None:
    provider_set = EnsembleSummaryProviderSet(
        {
            "First provider": EnsembleSummaryProviderMock("First mock"),
            "Second provider": EnsembleSummaryProviderMock("Second mock"),
            "Third provider": EnsembleSummaryProviderMock(("Third mock")),
        }
    )

    first_delta_ensemble = DeltaEnsemble(
        ensemble_a="First provider", ensemble_b="Third provider"
    )
    second_delta_ensemble = DeltaEnsemble(
        ensemble_a="Third provider", ensemble_b="Second provider"
    )

    first_provider_pair = create_delta_ensemble_provider_pair(
        first_delta_ensemble, provider_set
    )
    second_provider_pair = create_delta_ensemble_provider_pair(
        second_delta_ensemble, provider_set
    )

    assert len(first_provider_pair) == 2
    ensemble_a = first_provider_pair[0]
    ensemble_b = first_provider_pair[1]
    if not isinstance(ensemble_a, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected "{ensemble_a}" in second provider pair to be of type '
            "EnsembleSummaryProviderMock"
        )
    if not isinstance(ensemble_b, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected "{ensemble_b}" in second provider pair to be of type '
            "EnsembleSummaryProviderMock"
        )
    assert ensemble_a.get_name() == "First mock"
    assert ensemble_b.get_name() == "Third mock"

    assert len(second_provider_pair) == 2
    ensemble_a = second_provider_pair[0]
    ensemble_b = second_provider_pair[1]
    if not isinstance(ensemble_a, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected "{ensemble_a}" in second provider pair to be of type '
            "EnsembleSummaryProviderMock"
        )
    if not isinstance(ensemble_b, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected "{ensemble_b}" in second provider pair to be of type '
            "EnsembleSummaryProviderMock"
        )
    assert ensemble_a.get_name() == "Third mock"
    assert ensemble_b.get_name() == "Second mock"


def test_create_delta_ensemble_provider_pair_invalid_ensemble() -> None:
    provider_set = EnsembleSummaryProviderSet(
        {
            "First provider": EnsembleSummaryProviderMock("First mock"),
            "Second provider": EnsembleSummaryProviderMock("Second mock"),
        }
    )

    # Expect ValueError when retrieveing ensemble_b is invalid
    with pytest.raises(ValueError):
        invalid_provider = "Invalid provider"
        invalid_provider_pair = DeltaEnsemble(
            ensemble_a="First provider", ensemble_b=invalid_provider
        )
        create_delta_ensemble_provider_pair(invalid_provider_pair, provider_set)
        pytest.fail(
            "Expected ValueError when creating delta ensemble provider pair, as ensemble b = "
            f'"{invalid_provider}" is not existing in provider set!'
        )



from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from ..._models import GruptreeModel, WellAttributesModel
from ..._providers import (
    EnsembleSummaryProvider,
    EnsembleSummaryProviderFactory,
    Frequency,
)
from ._utils import EnsembleWellAnalysisData
from ._views._well_control_view import WellControlView, WellControlViewElement
from ._views._well_overview_view import WellOverviewView, WellOverviewViewElement


class WellAnalysis(WebvizPluginABC):
    """This plugin is for visualizing and analysing well data. There are different tabs
    for visualizing:

    * Well Production
    * Well control modes and network pressures

    ---
    * **`ensembles`:** Which ensembles in `shared_settings` to include.
    * **`rel_file_pattern`:** path to `.arrow` files with summary data.
    * **`gruptree_file`:** `.csv` with gruptree information.
    * **`time_index`:** Frequency for the data sampling.
    * **`filter_out_startswith`:** Filter out wells that starts with this string
    ---

    **Summary data**

    This plugin needs the following summary vectors to be exported:
    * WOPT, WGPT and WWPT for all wells for the well overview plots
    * WMCTL, WTHP and WBHP for all wells for the well control plots
    * GPR for all network nodes downstream/upstream the wells

    **GRUPTREE input**

    `gruptree_file` is a path to a file stored per realization (e.g. in \
    `share/results/tables/gruptree.csv"`).

    The `gruptree_file` file can be dumped to disk per realization by the `ECL2CSV` forward
    model with subcommand `gruptree`. The forward model uses `ecl2df` to export a table
    representation of the Eclipse network:
    [Link to ecl2csv gruptree documentation.](https://equinor.github.io/ecl2df/usage/gruptree.html).

    **time_index**

    This is the sampling interval of the summary data. It is `yearly` by default, but can be set
    to f.ex `monthly` if needed.

    **filter_out_startswith**

    Filter out well names that starts with this. Can f.ex be "R_" in order to filter out RFT wells
    without production.

    """

    class Ids(StrEnum):
        WELL_OVERVIEW = "well-overview"
        WELL_CONTROL = "well-control"

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: Optional[List[str]] = None,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        gruptree_file: str = "share/results/tables/gruptree.csv",
        well_attributes_file: str = "rms/output/wells/well_attributes.json",
        time_index: str = Frequency.YEARLY.value,
        filter_out_startswith: Optional[str] = None,
    ) -> None:
        super().__init__(stretch=True)

        self._ensembles = ensembles
        self._theme = webviz_settings.theme

        if ensembles is None:
            raise ValueError('Incorrect argument, must provide "ensembles"')

        self._ensemble_paths: Dict[str, Path] = {
            ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                ensemble_name
            ]
            for ensemble_name in ensembles
        }

        provider_factory = EnsembleSummaryProviderFactory.instance()

        self._data_models: Dict[str, EnsembleWellAnalysisData] = {}

        sampling = Frequency(time_index)
        for ens_name, ens_path in self._ensemble_paths.items():
            provider: EnsembleSummaryProvider = (
                provider_factory.create_from_arrow_unsmry_presampled(
                    str(ens_path), rel_file_pattern, sampling
                )
            )
            self._data_models[ens_name] = EnsembleWellAnalysisData(
                ens_name,
                provider,
                GruptreeModel(ens_name, ens_path, gruptree_file),
                WellAttributesModel(ens_name, ens_path, well_attributes_file),
                filter_out_startswith=filter_out_startswith,
            )

        self.add_view(
            WellOverviewView(self._data_models, self._theme),
            self.Ids.WELL_OVERVIEW,
        )
        self.add_view(
            WellControlView(self._data_models, self._theme),
            self.Ids.WELL_CONTROL,
        )

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.view(self.Ids.WELL_OVERVIEW)
                .view_element(WellOverviewView.Ids.VIEW_ELEMENT)
                .component_unique_id(WellOverviewViewElement.Ids.GRAPH),
                "content": "Shows production split on well for the various chart types",
            },
            {
                "id": self.view(self.Ids.WELL_OVERVIEW)
                .settings_group(WellOverviewView.Ids.CHART_TYPE)
                .get_unique_id(),
                "content": "Choose chart type: bar, pie or area chart. ",
            },
            {
                "id": self.view(self.Ids.WELL_OVERVIEW)
                .settings_group(WellOverviewView.Ids.SELECTIONS)
                .get_unique_id(),
                "content": "Choose ensembles, type of production and other options.",
            },
            {
                "id": self.view(self.Ids.WELL_OVERVIEW)
                .settings_group(WellOverviewView.Ids.LAYOUT_OPTIONS)
                .get_unique_id(),
                "content": "Chart layout options.",
            },
            {
                "id": self.view(self.Ids.WELL_OVERVIEW)
                .settings_group(WellOverviewView.Ids.FILTERS)
                .get_unique_id(),
                "content": "You can choose to view the production for all the wells or "
                "select only the ones you are interested in.",
            },
            {
                "id": self.view(self.Ids.WELL_CONTROL)
                .view_element(WellControlView.Ids.VIEW_ELEMENT)
                .component_unique_id(WellControlViewElement.Ids.CHART),
                "content": "Shows the number of realizations on different control modes "
                "and network pressures.",
            },
            {
                "id": self.view(self.Ids.WELL_CONTROL)
                .settings_group(WellControlView.Ids.SETTINGS)
                .get_unique_id(),
                "content": "Select the ensemble and well you are interested in."
                "The well dropdown is autmatically updated when ensemble is selected",
            },
            {
                "id": self.view(self.Ids.WELL_CONTROL)
                .settings_group(WellControlView.Ids.PRESSUREPLOT_OPTIONS)
                .get_unique_id(),
                "content": "Here are some options related only to the Network pressures plot.",
            },
        ]

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return [
            webviz_store_tuple
            for _, ens_data_model in self._data_models.items()
            for webviz_store_tuple in ens_data_model.webviz_store
        ]



from typing import Callable, Dict, List


def generate_tour_steps(get_uuid: Callable) -> List[Dict]:
    """Returns a list of reactour steps"""
    return [
        {
            "id": f"{get_uuid('layout')}",
            "content": ("Dashboard to analyze structural uncertainty from FMU results"),
        },
        {
            "id": f"{get_uuid('intersection-graph')}",
            "content": (
                "Intersection view displaying realizations or statistics of ensemble "
                "surfaces along a well trajectory or polyline. "
                "Zonation logs, if available, are represented as markers at each "
                "zone transition."
            ),
        },
        {
            "id": f"{get_uuid('intersection-source-wrapper')}",
            "content": (
                "Source of the intersection. Digitized polyline "
                "from a map or from a well trajectory (If wells are available)."
            ),
        },
        {
            "id": f"{get_uuid('all-maps-wrapper')}",
            "content": (
                "Map view displaying individiual surfaces. The left and center view "
                "are selecteable, while the right view displays the difference (left-center)."
                "Well markers for the displayed surface can be calculated on-the-fly and can "
                "be interacted with to update the intersection view."
            ),
        },
        {
            "id": f"{get_uuid('map-wrapper')} .leaflet-draw-draw-polyline",
            "content": (
                "Activate to draw a polyline on the map that will be displayed "
                "in the intersection view."
            ),
        },
        {
            "id": f"{get_uuid('intersection-data-wrapper')}",
            "content": ("Data used to populate the intersection view"),
        },
        {
            "id": f"{get_uuid('apply-intersection-data-selections')}",
            "content": (
                "Applies current changes " "and updates the intersection graph"
            ),
        },
        {
            "id": f"{get_uuid('surface-settings-wrapper')}",
            "content": (
                "Opens a panel for selecting data to display in " "the map views"
            ),
        },
        {
            "id": f"{get_uuid('realization-filter-wrapper')}",
            "content": (
                "Opens a panel to a global realization filter "
                "Statistics and available realizations in both "
                "the intersection view and the map views will "
                "use this filter."
            ),
        },
        {
            "id": f"{get_uuid('uncertainty-table-display-button')}",
            "content": (
                "Opens a panel to calculate surface intersection statistics "
                "for the active well (Not in use when no wells are available"
            ),
        },
    ]



from typing import Dict, List, TypedDict, Union

import pandas as pd
import webviz_core_components as wcc
from dash import html
from dash.development.base_component import Component
from webviz_config.common_cache import CACHE
from webviz_config.utils import StrEnum


class ColorBy(StrEnum):
    ENSEMBLE = "ENSEMBLE"
    PVTNUM = "PVTNUM"


class LayoutAttributes(TypedDict):
    x_axis_title: str
    y_axis_title: str
    df_column: str
    show_scatter_values: bool
    show_border_values: bool
    show_border_markers: bool


def filter_data_frame(
    data_frame: pd.DataFrame, ensembles: List[str], pvtnums: List[int]
) -> pd.DataFrame:

    data_frame = data_frame.copy()
    data_frame = data_frame.loc[data_frame["ENSEMBLE"].isin(ensembles)]
    data_frame = data_frame.loc[data_frame["PVTNUM"].isin(pvtnums)]
    return data_frame.fillna(0)


def create_hovertext(
    keyword: str,
    constant_group: str,
    group: str,
    color_by: ColorBy,
    realization: str,
    ratio_value: float,
) -> str:
    rs_v = (
        f"Rs = {ratio_value}, "
        if keyword == "PVTO"
        else f"Rv = {ratio_value}"
        if keyword == "PVTG"
        else ""
    )
    pvt_num = group if color_by == ColorBy.PVTNUM else constant_group
    ensemble = group if color_by == ColorBy.ENSEMBLE else constant_group

    return (
        f"{rs_v}Pvtnum: {pvt_num}<br />Realization: {realization}, Ensemble: {ensemble}"
    )


def create_traces(
    data_frame: pd.DataFrame,
    color_by: ColorBy,
    colors: Dict[str, List[str]],
    phase: str,
    column_name: str,
    show_scatter_values: bool,
    show_border_values: bool,
    show_border_markers: bool = False,
) -> List[dict]:
    """Renders line traces for individual realizations"""
    # pylint: disable-msg=too-many-locals
    # pylint: disable=too-many-nested-blocks
    # pylint: disable=too-many-branches

    traces = []
    hovertext: Union[List[str], str]
    border_value_pressure: Dict[str, list] = {}
    border_value_dependent: Dict[str, list] = {}

    dim_column_name = "RATIO"

    if phase == "OIL":
        data_frame = data_frame.loc[
            (data_frame["KEYWORD"] == "PVTO") | (data_frame["KEYWORD"] == "PVDO")
        ]
    elif phase == "GAS":
        data_frame = data_frame.loc[
            (data_frame["KEYWORD"] == "PVTG") | (data_frame["KEYWORD"] == "PVDG")
        ]
        dim_column_name = "PRESSURE"
    else:
        data_frame = data_frame.loc[data_frame["KEYWORD"] == "PVTW"]
        dim_column_name = "PRESSURE"

    data_frame = data_frame.sort_values(
        ["PRESSURE", "VOLUMEFACTOR", "VISCOSITY"],
        ascending=[True, True, True],
    )

    constant_group = (
        data_frame["PVTNUM"].iloc[0]
        if color_by == ColorBy.ENSEMBLE
        else data_frame["ENSEMBLE"].iloc[0]
    )

    for (group, grouped_data_frame) in data_frame.groupby(color_by):
        for set_no, set_value in enumerate(
            grouped_data_frame[dim_column_name].unique()
        ):
            for realization_no, (realization, realization_data_frame) in enumerate(
                grouped_data_frame.groupby("REAL")
            ):
                if group not in border_value_pressure:
                    border_value_pressure[group] = []
                    border_value_dependent[group] = []
                try:
                    border_value_pressure[group].append(
                        realization_data_frame.loc[
                            realization_data_frame[dim_column_name] == set_value
                        ]["PRESSURE"].iloc[0]
                    )
                    if column_name == "VISCOSITY":
                        if phase == "OIL":
                            border_value_dependent[group].append(
                                realization_data_frame[
                                    (
                                        realization_data_frame[dim_column_name]
                                        == set_value
                                    )
                                ]["VISCOSITY"].iloc[0]
                            )
                        else:
                            border_value_dependent[group].append(
                                realization_data_frame[
                                    (
                                        realization_data_frame[dim_column_name]
                                        == set_value
                                    )
                                ]["VISCOSITY"].max()
                            )
                    else:
                        border_value_dependent[group].append(
                            realization_data_frame.loc[
                                realization_data_frame[dim_column_name] == set_value
                            ][column_name].iloc[0]
                        )
                except IndexError as exc:
                    raise IndexError(
                        "This error is most likely due to PVT differences between "
                        "realizations within the same ensemble. This is currently not "
                        "supported."
                    ) from exc

                if show_scatter_values:
                    if phase == "GAS":
                        hovertext = [
                            create_hovertext(
                                realization_data_frame["KEYWORD"].iloc[0],
                                constant_group,
                                group,
                                color_by,
                                realization,
                                realization_data_frame.loc[
                                    (realization_data_frame["PRESSURE"] == y)
                                    & (realization_data_frame[column_name] == x)
                                ]["RATIO"].iloc[0],
                            )
                            for x, y in zip(
                                realization_data_frame.loc[
                                    realization_data_frame["PRESSURE"] == set_value
                                ][column_name],
                                realization_data_frame.loc[
                                    realization_data_frame["PRESSURE"] == set_value
                                ].PRESSURE,
                            )
                        ]
                    else:
                        hovertext = create_hovertext(
                            realization_data_frame["KEYWORD"].iloc[0],
                            constant_group,
                            group,
                            color_by,
                            realization,
                            set_value,
                        )

                    traces.extend(
                        [
                            {
                                "type": "scatter",
                                "x": realization_data_frame.loc[
                                    realization_data_frame[dim_column_name] == set_value
                                ]["PRESSURE"],
                                "y": realization_data_frame.loc[
                                    realization_data_frame[dim_column_name] == set_value
                                ][column_name],
                                "xaxis": "x",
                                "yaxis": "y",
                                "hovertext": hovertext,
                                "name": group,
                                "legendgroup": group,
                                "marker": {
                                    "color": colors.get(
                                        group, colors[list(colors.keys())[-1]]
                                    )
                                },
                                "showlegend": realization_no == 0 and set_no == 0,
                            }
                        ]
                    )
    if show_border_values:
        for group, group_border_value_pressure in border_value_pressure.items():
            traces.extend(
                [
                    {
                        "type": "scatter",
                        "mode": ("lines+markers" if show_border_markers else "lines"),
                        "x": group_border_value_pressure,
                        "y": border_value_dependent[group],
                        "xaxis": "x",
                        "yaxis": "y",
                        "legendgroup": group,
                        "line": {
                            "width": 1,
                            "color": colors.get(group, colors[list(colors.keys())[-1]]),
                        },
                        "showlegend": not show_scatter_values,
                    }
                ]
            )
    return traces


def create_graph(
    data_frame: pd.DataFrame,
    color_by: ColorBy,
    colors: Dict[str, List[str]],
    phase: str,
    plot: str,
    plot_title: str,
    theme: dict,
    graph_height: float,
) -> Component:
    layout_attributes: Dict[str, LayoutAttributes] = (
        {
            "fvf": {
                "x_axis_title": rf"Pressure [{data_frame['PRESSURE_UNIT'].iloc[0]}]",
                "y_axis_title": rf"[{data_frame['VOLUMEFACTOR_UNIT'].iloc[0]}]",
                "df_column": "VOLUMEFACTOR",
                "show_scatter_values": True,
                "show_border_values": True,
                "show_border_markers": False,
            },
            "viscosity": {
                "x_axis_title": rf"Pressure [{data_frame['PRESSURE_UNIT'].iloc[0]}]",
                "y_axis_title": rf"[{data_frame['VISCOSITY_UNIT'].iloc[0]}]",
                "df_column": "VISCOSITY",
                "show_scatter_values": True,
                "show_border_values": True,
                "show_border_markers": False,
            },
            "density": {
                "x_axis_title": rf"Pressure [{data_frame['PRESSURE_UNIT'].iloc[0]}]",
                "y_axis_title": rf"[{data_frame['DENSITY_UNIT'].iloc[0]}]",
                "df_column": "DENSITY",
                "show_scatter_values": True,
                "show_border_values": True,
                "show_border_markers": False,
            },
            "ratio": {
                "x_axis_title": rf"Pressure [{data_frame['PRESSURE_UNIT'].iloc[0]}]",
                "y_axis_title": rf"[{data_frame['RATIO_UNIT'].iloc[0]}]",
                "df_column": "DENSITY",
                "show_scatter_values": False,
                "show_border_values": True,
                "show_border_markers": True,
            },
        }
        if not data_frame.empty
        else {}
    )

    return wcc.FlexBox(
        style={
            "height": f"{graph_height}vh",
            "min-width": "20vh",
            "text-align": "center",
            "padding": "2%",
            # "flex": "1 1 46%",
        },
        children=(
            [
                html.Span(plot_title, style={"font-weight": "bold"}),
                wcc.Graph(
                    style={"height": f"{graph_height}vh"},
                    figure={
                        "layout": plot_layout(
                            color_by,
                            theme,
                            layout_attributes[plot]["x_axis_title"],
                            layout_attributes[plot]["y_axis_title"],
                        ),
                        "data": create_traces(
                            data_frame,
                            color_by,
                            colors,
                            phase,
                            layout_attributes[plot]["df_column"],
                            layout_attributes[plot]["show_scatter_values"],
                            layout_attributes[plot]["show_border_values"],
                            layout_attributes[plot]["show_border_markers"],
                        ),
                    }
                    if not data_frame.empty and plot in layout_attributes
                    else {},
                ),
            ]
        ),
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def plot_layout(
    color_by: ColorBy,
    theme: dict,
    x_unit: str,
    y_unit: str,
) -> dict:
    layout = {}
    layout.update(theme)
    layout["legend"] = {"title": {"text": color_by.lower().capitalize()}}
    layout.update(
        {
            "xaxis": {
                "automargin": True,
                "zeroline": False,
                "anchor": "y",
                "domain": [0.0, 1.0],
                "title": {
                    "text": x_unit,
                    "standoff": 15,
                },
                "showticklabels": True,
                "showgrid": True,
            },
            "yaxis": {
                "automargin": True,
                "ticks": "",
                "zeroline": False,
                "anchor": "x",
                "domain": [0.0, 1.0],
                "title": {
                    "text": y_unit,
                },
                "type": "linear",
                "showgrid": True,
            },
            "margin": {"t": 20, "b": 0},
            "plot_bgcolor": "rgba(0,0,0,0)",
            "hovermode": "closest",
        }
    )
    return layout



from typing import List

import pandas as pd
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ......_components.parameter_filter import ParameterFilter


class ParameterFilterSettings(SettingsGroupABC):
    class Ids(StrEnum):
        PARAM_FILTER = "param-filter"

    def __init__(self, parameter_df: pd.DataFrame, mc_ensembles: List[str]) -> None:
        super().__init__("Parameter Filter")
        self._parameter_df = parameter_df
        self._mc_ensembles = mc_ensembles

    def layout(self) -> List[Component]:
        return ParameterFilter(
            uuid=self.register_component_unique_id(self.Ids.PARAM_FILTER),
            dframe=self._parameter_df[
                self._parameter_df["ENSEMBLE"].isin(self._mc_ensembles)
            ].copy(),
            reset_on_ensemble_update=True,
            display_header=False,
        ).layout



from typing import Callable, Optional

from dash import MATCH, Dash, Input, Output, State
from dash.exceptions import PreventUpdate


def open_dialogs(
    app: Dash,
    get_uuid: Callable,
) -> None:
    @app.callback(
        Output(
            {"id": get_uuid("dialog"), "dialog_id": MATCH, "element": "wrapper"},
            "open",
        ),
        Input(
            {"id": get_uuid("dialog"), "dialog_id": MATCH, "element": "button-open"},
            "n_clicks",
        ),
        State(
            {"id": get_uuid("dialog"), "dialog_id": MATCH, "element": "wrapper"},
            "open",
        ),
    )
    def _toggle_dialog_graph_settings(n_open: int, is_open: bool) -> Optional[bool]:
        """Open or close graph settings dialog button"""
        if n_open:
            return not is_open
        raise PreventUpdate



import datetime

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal
from webviz_subsurface_components.VectorCalculatorWrapper import (
    ExpressionInfo,
    VariableVectorMapInfo,
)

from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object

# pylint: disable = line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.derived_vectors_accessor.derived_delta_ensemble_vectors_accessor_impl import (
    DerivedDeltaEnsembleVectorsAccessorImpl,
)

from ..mocks.derived_vectors_accessor_ensemble_summary_provider_mock import (
    EnsembleSummaryProviderMock,
)

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************

# fmt: off
# Ensemble A
INPUT_A_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "A", "B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 10.0,   500.0  ],
        [datetime.datetime(2000,2,1),  1, 20.0,   1000.0 ],
        [datetime.datetime(2000,3,1),  1, 30.0,   1500.0 ],
        [datetime.datetime(2000,1,1),  2, 60.0,   3000.0 ],
        [datetime.datetime(2000,2,1),  2, 70.0,   4000.0 ],
        [datetime.datetime(2000,3,1),  2, 80.0,   5000.0 ],
        [datetime.datetime(2000,1,1),  4, 110.0,  10000.0],
        [datetime.datetime(2000,2,1),  4, 120.0,  12000.0],
        [datetime.datetime(2000,3,1),  4, 130.0,  14000.0],
    ]
)

# Ensemble B
INPUT_B_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "A", "B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 1.0,   50.0  ],
        [datetime.datetime(2000,2,1),  1, 2.0,   100.0 ],
        [datetime.datetime(2000,3,1),  1, 3.0,   150.0 ],
        [datetime.datetime(2000,1,1),  2, 6.0,   300.0 ],
        [datetime.datetime(2000,2,1),  2, 7.0,   400.0 ],
        [datetime.datetime(2000,3,1),  2, 8.0,   500.0 ],
        [datetime.datetime(2000,1,1),  4, 11.0,  1000.0],
        [datetime.datetime(2000,2,1),  4, 12.0,  1200.0],
        [datetime.datetime(2000,3,1),  4, 13.0,  1400.0],
    ]
)

# Delta between A and B DF (A-B)
EXPECTED_DELTA_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "A", "B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 9.0,    450.0  ],
        [datetime.datetime(2000,2,1),  1, 18.0,   900.0  ],
        [datetime.datetime(2000,3,1),  1, 27.0,   1350.0 ],
        [datetime.datetime(2000,1,1),  2, 54.0,   2700.0 ],
        [datetime.datetime(2000,2,1),  2, 63.0,   3600.0 ],
        [datetime.datetime(2000,3,1),  2, 72.0,   4500.0 ],
        [datetime.datetime(2000,1,1),  4, 99.0,   9000.0 ],
        [datetime.datetime(2000,2,1),  4, 108.0,  10800.0],
        [datetime.datetime(2000,3,1),  4, 117.0,  12600.0],
    ]
)

# PER_INTVL_ calc for col "B" of Delta
EXPECTED_DELTA_INVTL_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "PER_INTVL_B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 450.0 ],
        [datetime.datetime(2000,2,1),  1, 450.0 ],
        [datetime.datetime(2000,3,1),  1, 0.0   ],
        [datetime.datetime(2000,1,1),  2, 900.0 ],
        [datetime.datetime(2000,2,1),  2, 900.0 ],
        [datetime.datetime(2000,3,1),  2, 0.0   ],
        [datetime.datetime(2000,1,1),  4, 1800.0],
        [datetime.datetime(2000,2,1),  4, 1800.0],
        [datetime.datetime(2000,3,1),  4, 0.0   ],
    ]
)

# Sum of col "A" and "B" in Delta
EXPECTED_SUM_A_AND_B_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "Sum A and B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 459.0  ],
        [datetime.datetime(2000,2,1),  1, 918.0  ],
        [datetime.datetime(2000,3,1),  1, 1377.0 ],
        [datetime.datetime(2000,1,1),  2, 2754.0 ],
        [datetime.datetime(2000,2,1),  2, 3663.0 ],
        [datetime.datetime(2000,3,1),  2, 4572.0 ],
        [datetime.datetime(2000,1,1),  4, 9099.0 ],
        [datetime.datetime(2000,2,1),  4, 10908.0],
        [datetime.datetime(2000,3,1),  4, 12717.0],
    ]
)
make_date_column_datetime_object(INPUT_A_DF)
make_date_column_datetime_object(INPUT_B_DF)
make_date_column_datetime_object(EXPECTED_DELTA_DF)
make_date_column_datetime_object(EXPECTED_DELTA_INVTL_DF)
make_date_column_datetime_object(EXPECTED_SUM_A_AND_B_DF)


# Dates AFTER year 2262!
AFTER_2262_DATES = pd.Series(
    [
        datetime.datetime(2265,1,1),
        datetime.datetime(2265,2,1),
        datetime.datetime(2265,3,1),
        datetime.datetime(2265,1,1),
        datetime.datetime(2265,2,1),
        datetime.datetime(2265,3,1),
        datetime.datetime(2265,1,1),
        datetime.datetime(2265,2,1),
        datetime.datetime(2265,3,1),
    ]
)
# NOTE: datetime.datetime after year 2262 is not converted to pd.Timestamp, thus
# no need to make date column datetime object
INPUT_A_AFTER_2262_DF = INPUT_A_DF.copy()
INPUT_A_AFTER_2262_DF["DATE"] = AFTER_2262_DATES
INPUT_B_AFTER_2262_DF = INPUT_B_DF.copy()
INPUT_B_AFTER_2262_DF["DATE"] = AFTER_2262_DATES
EXPECTED_DELTA_AFTER_2262_DF = EXPECTED_DELTA_DF.copy()
EXPECTED_DELTA_AFTER_2262_DF["DATE"] = AFTER_2262_DATES
EXPECTED_DELTA_INVTL_AFTER_2262_DF = EXPECTED_DELTA_INVTL_DF.copy()
EXPECTED_DELTA_INVTL_AFTER_2262_DF["DATE"] = AFTER_2262_DATES
EXPECTED_SUM_A_AND_B_AFTER_2262_DF = EXPECTED_SUM_A_AND_B_DF.copy()
EXPECTED_SUM_A_AND_B_AFTER_2262_DF["DATE"] = AFTER_2262_DATES

# fmt: on
TEST_EXPRESSION = ExpressionInfo(
    name="Sum A and B",
    expression="x+y",
    id="TestId",
    variableVectorMap=[
        VariableVectorMapInfo(variableName="x", vectorName=["A"]),
        VariableVectorMapInfo(variableName="y", vectorName=["B"]),
    ],
    isValid=True,
    isDeletable=False,
)

TEST_ACCESSOR = DerivedDeltaEnsembleVectorsAccessorImpl(
    name="Test accessor",
    provider_pair=(
        EnsembleSummaryProviderMock(INPUT_A_DF),
        EnsembleSummaryProviderMock(INPUT_B_DF),
    ),
    vectors=["A", "B", "PER_INTVL_B", "Sum A and B"],
    expressions=[TEST_EXPRESSION],
    resampling_frequency=None,
)

TEST_AFTER_2262_ACCESSOR = DerivedDeltaEnsembleVectorsAccessorImpl(
    name="Test 2262 accessor",
    provider_pair=(
        EnsembleSummaryProviderMock(INPUT_A_AFTER_2262_DF),
        EnsembleSummaryProviderMock(INPUT_B_AFTER_2262_DF),
    ),
    vectors=["A", "B", "PER_INTVL_B", "Sum A and B"],
    expressions=[TEST_EXPRESSION],
    resampling_frequency=None,
)

TEST_EMPTY_ACCESSOR = DerivedDeltaEnsembleVectorsAccessorImpl(
    name="Empty provider accessor",
    provider_pair=(
        EnsembleSummaryProviderMock(pd.DataFrame()),
        EnsembleSummaryProviderMock(pd.DataFrame()),
    ),
    vectors=["A", "B", "PER_INTVL_B", "Sum A and B"],
    expressions=None,
    resampling_frequency=None,
)

# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************

TEST_STATUS_CASES = [
    pytest.param(TEST_ACCESSOR, True),
    pytest.param(TEST_AFTER_2262_ACCESSOR, True),
    pytest.param(TEST_EMPTY_ACCESSOR, False),
]
TEST_GET_VECTOR_CASES = [
    pytest.param(TEST_ACCESSOR, EXPECTED_DELTA_DF),
    pytest.param(TEST_AFTER_2262_ACCESSOR, EXPECTED_DELTA_AFTER_2262_DF),
]
TEST_CREATE_PER_INTVL_PER_DAY_VECTOR_CASES = [
    pytest.param(TEST_ACCESSOR, EXPECTED_DELTA_INVTL_DF),
    pytest.param(TEST_AFTER_2262_ACCESSOR, EXPECTED_DELTA_INVTL_AFTER_2262_DF),
]
TEST_CREATE_CALCULATED_VECTOR_CASES = [
    pytest.param(TEST_ACCESSOR, EXPECTED_SUM_A_AND_B_DF),
    pytest.param(TEST_AFTER_2262_ACCESSOR, EXPECTED_SUM_A_AND_B_AFTER_2262_DF),
]


@pytest.mark.parametrize("test_accessor, expected_state", TEST_STATUS_CASES)
def test_has_provider_vectors(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_state: bool
) -> None:
    assert test_accessor.has_provider_vectors() == expected_state


@pytest.mark.parametrize("test_accessor, expected_state", TEST_STATUS_CASES)
def test_has_per_interval_and_per_day_vectors(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_state: bool
) -> None:
    assert test_accessor.has_per_interval_and_per_day_vectors() == expected_state


@pytest.mark.parametrize("test_accessor, expected_state", TEST_STATUS_CASES)
def test_has_vector_calculator_expressions(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_state: bool
) -> None:
    assert test_accessor.has_vector_calculator_expressions() == expected_state


@pytest.mark.parametrize("test_accessor, expected_df", TEST_GET_VECTOR_CASES)
def test_get_provider_vectors(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    assert_frame_equal(expected_df, test_accessor.get_provider_vectors_df())


@pytest.mark.parametrize("test_accessor, expected_df", TEST_GET_VECTOR_CASES)
def test_get_provider_vectors_filter_realizations(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    # Filter realizations
    expected_reals_df = (
        expected_df.loc[expected_df["REAL"].isin([2, 4])]
        .reset_index()
        .drop("index", axis=1)
    )

    test_df = test_accessor.get_provider_vectors_df(realizations=[2, 4])

    assert_frame_equal(expected_reals_df, test_df)
    assert list(set(test_df["REAL"].values)) == [2, 4]


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_PER_INTVL_PER_DAY_VECTOR_CASES
)
def test_create_per_interval_and_per_day_vectors_df(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    assert_frame_equal(
        expected_df, test_accessor.create_per_interval_and_per_day_vectors_df()
    )


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_PER_INTVL_PER_DAY_VECTOR_CASES
)
def test_create_per_interval_and_per_day_vectors_df_filter_realizations(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    # Filter realizations
    expected_reals_df = (
        expected_df.loc[expected_df["REAL"].isin([1, 4])]
        .reset_index()
        .drop("index", axis=1)
    )

    test_df = test_accessor.create_per_interval_and_per_day_vectors_df(
        realizations=[1, 4]
    )

    assert_frame_equal(expected_reals_df, test_df)
    assert list(set(test_df["REAL"].values)) == [1, 4]


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_CALCULATED_VECTOR_CASES
)
def test_create_calculated_vectors_df(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    assert_frame_equal(expected_df, test_accessor.create_calculated_vectors_df())


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_CALCULATED_VECTOR_CASES
)
def test_create_calculated_vectors_df_filter_realizations(
    test_accessor: DerivedDeltaEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    # Filter realizations
    expected_reals_df = (
        expected_df.loc[expected_df["REAL"].isin([1, 2])]
        .reset_index()
        .drop("index", axis=1)
    )

    test_df = test_accessor.create_calculated_vectors_df(realizations=[1, 2])

    assert_frame_equal(expected_reals_df, test_df)
    assert list(set(test_df["REAL"].values)) == [1, 2]



import itertools
from typing import List, Tuple

import dash


def update_relevant_components(id_list: list, update_info: List[dict]) -> list:
    output_id_list = [dash.no_update] * len(id_list)
    for elm in update_info:
        for idx, x in enumerate(id_list):
            if all(x[key] == value for key, value in elm["conditions"].items()):
                output_id_list[idx] = elm["new_value"]
                break
    return output_id_list


def move_to_end_of_list(element: str, list_of_elements: list) -> list:
    return [x for x in list_of_elements if x != element] + [element]


def to_ranges(list_of_integers: List[int]) -> List[Tuple[int, int]]:
    """Return list of tuples with ranges from list of integers"""
    ranges = []
    for _, group in itertools.groupby(
        enumerate(sorted(list_of_integers)), lambda t: t[1] - t[0]
    ):
        int_range = list(group)
        ranges.append((int_range[0][1], int_range[-1][1]))
    return ranges



import glob
import warnings
from pathlib import Path
from typing import List, Optional, Union

import pandas as pd
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

# The fmu.ensemble dependency ecl is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from fmu.ensemble import EnsembleSet, ScratchEnsemble
except ImportError:
    pass


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def scratch_ensemble(
    ensemble_name: str, ensemble_path: Path, filter_file: Union[str, None] = "OK"
) -> ScratchEnsemble:
    return (
        ScratchEnsemble(ensemble_name, ensemble_path)
        if filter_file is None
        else ScratchEnsemble(ensemble_name, ensemble_path).filter(filter_file)
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def load_ensemble_set(
    ensemble_paths: dict,
    ensemble_set_name: str = "EnsembleSet",
    filter_file: Union[str, None] = "OK",
) -> EnsembleSet:
    return EnsembleSet(
        ensemble_set_name,
        [
            scratch_ensemble(ens_name, ens_path, filter_file)
            for ens_name, ens_path in ensemble_paths.items()
        ],
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_parameters(
    ensemble_paths: dict,
    ensemble_set_name: str = "EnsembleSet",
    filter_file: Union[str, None] = "OK",
) -> pd.DataFrame:
    return load_ensemble_set(ensemble_paths, ensemble_set_name, filter_file).parameters


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_csv(
    ensemble_paths: dict, csv_file: str, ensemble_set_name: str = "EnsembleSet"
) -> pd.DataFrame:

    return load_ensemble_set(ensemble_paths, ensemble_set_name).load_csv(csv_file)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_smry(
    ensemble_paths: dict,
    ensemble_set_name: str = "EnsembleSet",
    time_index: Optional[Union[list, str]] = None,
    column_keys: Optional[list] = None,
) -> pd.DataFrame:

    return load_ensemble_set(ensemble_paths, ensemble_set_name).get_smry(
        time_index=time_index, column_keys=column_keys
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_smry_meta(
    ensemble_paths: dict,
    ensemble_set_name: str = "EnsembleSet",
    column_keys: Optional[list] = None,
) -> pd.DataFrame:
    """Finds metadata for the summary vectors in the ensemble set.
    Note that we assume the same units for all ensembles.
    (meaning that we update/overwrite when checking the next ensemble)
    """
    ensemble_set = load_ensemble_set(ensemble_paths, ensemble_set_name)
    smry_meta = {}
    for ensname in ensemble_set.ensemblenames:
        smry_meta.update(ensemble_set[ensname].get_smry_meta(column_keys=column_keys))
    return pd.DataFrame(smry_meta).transpose()


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def get_realizations(
    ensemble_paths: dict, ensemble_set_name: str = "EnsembleSet"
) -> pd.DataFrame:
    """Extracts realization info from a fmu.ensemble.EnsembleSet
    The information extracted is the ensemble name, realization number,
    realization local runpath, sensitivity name, sensitivity case and sensitivity type.
    The sensitivtiy information is only relevant if a design matrix is used. If the ensemble
    is a full monte carlo / history matching run this information will be undefined.

    Returns a pandas dataframe with columns: ENSEMBLE, REAL, RUNPATH, SENSNAME, SENSCASE, SENSTYPE
    """
    ens_set = load_ensemble_set(ensemble_paths, ensemble_set_name)
    df = ens_set.parameters.get(["ENSEMBLE", "REAL"])
    df["SENSCASE"] = ens_set.parameters.get("SENSCASE")
    df["SENSNAME"] = ens_set.parameters.get("SENSNAME")
    df["SENSTYPE"] = df.apply(lambda row: find_sens_type(row.SENSCASE), axis=1)
    df["RUNPATH"] = df.apply(
        # Extracts realization runpath from the EnsembleSet.ScratchEnsemble.Realization object
        lambda x: ens_set[x["ENSEMBLE"]][x["REAL"]].runpath(),
        axis=1,
    )
    return df.sort_values(by=["ENSEMBLE", "REAL"])


def find_sens_type(senscase: str) -> Optional[str]:
    """Finds sensitivity type from sensitivty case.
    If sensitivity case is 'p10_p90', sensitivity type is montecarlo,
    else sensitivity type is set to 'scalar'.
    """
    if not senscase:
        return None

    if senscase == "p10_p90":
        return "mc"

    return "scalar"


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def find_surfaces(
    ensemble_paths: dict,
    surface_folder: str = "share/results/maps",
    surface_files: Optional[List] = None,
    suffix: str = "*.gri",
    delimiter: str = "--",
) -> pd.DataFrame:
    """Reads surface file names stored in standard FMU format, and returns a dictionary
    on the following format:
    surface_property:
        names:
            - some_surface_name
            - another_surface_name
        dates:
            - some_date
            - another_date
    """
    # Create list of all files in all realizations in all ensembles
    files = []
    for _, ensdf in get_realizations(ensemble_paths=ensemble_paths).groupby("ENSEMBLE"):
        ens_files = []
        for _real_no, realdf in ensdf.groupby("REAL"):
            runpath = realdf.iloc[0]["RUNPATH"]
            for realpath in glob.glob(str(Path(runpath) / surface_folder / suffix)):
                filename = Path(realpath)
                if surface_files and filename.name not in surface_files:
                    continue
                stem = filename.stem.split(delimiter)
                if len(stem) >= 2:
                    ens_files.append(
                        {
                            "path": realpath,
                            "name": stem[0],
                            "attribute": stem[1],
                            "date": stem[2] if len(stem) >= 3 else None,
                            **realdf.iloc[0],
                        }
                    )
        if not ens_files:
            warnings.warn(f"No surfaces found for ensemble located at {runpath}.")
        else:
            files.extend(ens_files)

    # Store surface name, attribute and date as Pandas dataframe
    if not files:
        raise ValueError(
            "No surfaces found! Ensure that surfaces file are stored "
            "at share/results/maps in each ensemble and is following "
            "the FMU naming standard (name--attribute[--date].gri)"
        )
    return pd.DataFrame(files)



from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from ._utils._rft_plotter_data_model import RftPlotterDataModel
from ._views._map_view import MapView
from ._views._misfit_per_real_view import MisfitPerRealView
from ._views._parameter_response_view import ParameterResponseView
from ._views._sim_vs_obs_view import SimVsObsView


class RftPlotter(WebvizPluginABC):
    """This plugin visualizes simulated RFT results from
FMU ensembles combined with ERT observation data.

Several visualizations are available:

* Map view of RFT observations.

* Depth vs pressure plot showing simulated RFT data along wells together with observation points.

* Barchart showing sum of mean misfit for ERT observations per realization. One plot per ensemble.

* Crossplot of simulated RFT vs observed value per ERT observation. One plot per ensemble.

* Boxplot showing misfit per ERT observation for each ensemble.

* Plots for analysing the parameter response of the RFT data. Not available for aggregated data.

---
**Using data per realization**

* **`ensembles`**: Which ensembles in `shared_settings` to visualize.

In addition, you need to have rft-files in your realizations stored at the local path \
`share/results/tables`. The `rft_ert.csv` is required as input, while the `rft.csv` is optional:

* **`rft_ert.csv`**: A csv file containing simulated and observed RFT data for RFT observations \
defined in ERT \
[(example file)](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
01_drogon_ahm/realization-0/iter-0/share/results/tables/rft_ert.csv).

* **`rft.csv`**: A csv file containing simulated RFT data extracted from ECLIPSE RFT output files \
using [ecl2df](https://equinor.github.io/ecl2df/ecl2df.html#module-ecl2df.rft) \
[(example file)](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
01_drogon_ahm/realization-0/iter-0/share/results/tables/rft.csv). \
Simulated RFT data can be visualized along MD if a "CONMD" column is present in \
the dataframe and only for wells where each RFT datapoint has a unique MD.

* **`parameters.txt`**: File with parameters and values

**Using aggregated data**

* **`csvfile_rft`**: Aggregated version of `rft.csv` [(example file)](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_test_data/aggregated_data/rft.csv).
* **`csvfile_rft_ert`**: Aggregated version of `rft_ert.csv` [(example file)](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_test_data/aggregated_data/rft_ert.csv).


**Optional input for both input options**

* **`obsdata`**: A csv file containing additional RFT observation data not defined in ERT for
visualization together with simulated RFT.
Mandatory column names: `WELL`, `DATE` (yyyy-mm-dd), `DEPTH` and `PRESSURE`

* **`formations`**: A csv file containing top and base values for each zone per well.
Used to visualize zone boundaries together with simulated RFT.
Mandatory column names: `WELL`, `ZONE`, `TOP_TVD`, `BASE_TVD` \
[(example file))](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
01_drogon_ahm/realization-0/iter-0/share/results/tables/formations.csv).

* **`faultlines`**: A csv file containing faultpolygons to be visualized together with the map view.
Export format from [xtgeo.xyz.polygons.dataframe](
https://xtgeo.readthedocs.io/en/latest/apiref/xtgeo.xyz.polygons.html#xtgeo.xyz.polygons.Polygons.dataframe
) \
[(example file)](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/01_drogon_ahm/\
realization-0/iter-0/share/results/polygons/toptherys--gl_faultlines_extract_postprocess.csv).

---
?> Well name needs to be consistent with Eclipse well name.

?> Only RFT observations marked as active in ERT are used to generate plots.

The `rft_ert.csv` file can be generated by running the "MERGE_RFT_ERTOBS" forward model in ERT, \
this will collect ERT RFT observations and merge with CSV output from the "GENDATA_RFT" forward \
model. [ERT docs](https://fmu-docs.equinor.com/docs/ert/reference/\
forward_models.html?highlight=gendata_rft#MERGE_RFT_ERTOBS).

"""

    class Ids(StrEnum):
        MAP_VIEW = "map-view"
        MISFIT_PER_REAL_VIEW = "misfit-per-real-view"
        SIM_VS_OBS_VIEW = "sim-vs-obs-view"
        PARAMETER_RESPONSE_VIEW = "parameter-response-view"

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        csvfile_rft: Path = None,
        csvfile_rft_ert: Path = None,
        ensembles: Optional[List[str]] = None,
        formations: Path = None,
        obsdata: Path = None,
        faultlines: Path = None,
    ) -> None:
        super().__init__()

        self._datamodel = RftPlotterDataModel(
            webviz_settings,
            ensembles,
            formations,
            faultlines,
            obsdata,
            csvfile_rft,
            csvfile_rft_ert,
        )

        self.add_view(MapView(self._datamodel), self.Ids.MAP_VIEW)
        self.add_view(MisfitPerRealView(self._datamodel), self.Ids.MISFIT_PER_REAL_VIEW)
        self.add_view(SimVsObsView(self._datamodel), self.Ids.SIM_VS_OBS_VIEW)
        if not self._datamodel.param_model.sensrun:
            self.add_view(
                ParameterResponseView(self._datamodel), self.Ids.PARAMETER_RESPONSE_VIEW
            )

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return self._datamodel.webviz_store
